/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus;

import javax.swing.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.HashMap;

/**
 * This class manages the windows. As the usercan open as many windows as he wants in a
 * single instance of Klaus, these windows have to be managed.
 * @author Stephan Hillebrand
 * @since 2009-03-10
 */
public class KlausWindowManager {
	/**
	 * This HashMap stores all windows. Identification is done by a String handle.
	 */
	private HashMap<String, KlausWindow> windows = new HashMap<String, KlausWindow>();
	
	private static KlausWindowManager manager = new KlausWindowManager();
	private static WindowManagerThread updateThread = null;
	
	public static KlausWindowManager getInstance(){
		return manager;
	}
	
	
	public KlausWindowManager(){

	}
	
	/**
	 * Start update thread. This has to be done AFTER the first window (main window)
	 * was opened. Else the program might terminate before the window was opened on
	 * rather slow machines.
	 *
	 */
	public void startWindowManager(){
		updateThread = new WindowManagerThread();
		updateThread.start();
	}
	
	public KlausWindow getWindowByHandle(final String handle){
		KlausWindow window = windows.get(handle);
		if(window == null){
			//If the window doesn't exist, it obviously has to be created.
			//(What a surprise!)
			window = new KlausWindow();
			window.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
			windows.put(handle, window);
			
			window.addWindowListener(new WindowAdapter(){
				@Override
				@SuppressWarnings("unused")
				public void windowClosing(WindowEvent event){
					System.out.println("Window "+handle+" closed!");
					if(windows.get(handle).closeAllTabs()){
						windows.get(handle).setVisible(false);
						windows.remove(handle);
					}
					
				}
			});
		}
		return window;
	}
	
	/**
	 * This thread is necessary for periodical updates of the window list.
	 * Up to now, it only checks which windows are closed. These windows are
	 * removed. If no windows are opened, the program will exit.
	 * @author Stephan Hillebrand
	 * @since 2009-03-11
	 */
	private class WindowManagerThread extends Thread {
		/**
		 * Counts the number of updates without any opened windows.
		 * If this ever reaches 5, the program will exit. Whenever a
		 * new window is found, this is set to 0 again.
		 * This will reduce the propability of a crash to almost 0.
		 */
		private int emptyUpdates = 0;
		
		@Override
		public void run(){
			while(true){
				//Get a list of all Klaus windows.
				KlausWindow[] list = windows.values().toArray(new KlausWindow[]{});
				//A copy of the keys is necessary for reasons of thread safety
				String[] keys = windows.keySet().toArray(new String[]{});
				if(list.length == 0){
					++emptyUpdates;
					if(emptyUpdates > 1){
						//No windows have been opened for 5 seconds
						System.out.println("KlausWindowManager$WindowManagerThread#run(): Exit!");
						//Save config
						Settings.getSettings().save();
						System.exit(0);
					}
				}else
					emptyUpdates = 0;
				
				for(int i = 0; i < list.length; i++){
					if(!list[i].isVisible()){
						//Window i is closed, drop it!
						windows.remove(keys[i]);
					}
				}
				
				try{
					Thread.sleep(1000);
					//One sync every second should be enough.
				}catch(Exception e){}
			}
		}
	}
}
