/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus;

import klaus.engine.EngineData;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import javax.swing.*;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.text.MessageFormat;
import static java.text.MessageFormat.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

public class Settings {
	
	private static Settings settings = new Settings();
	
	private ArrayList<EngineData> engines = new ArrayList<EngineData>();
	private int defaultEngine = -1;
	
	public static Settings getSettings(){		
		return settings;
	}
	
	private String layoutName = "Woody";
	private String username = "";

	private String clockMode = "DIGITAL";
	
	private String lastICSUser = "guest", lastICSPWD = "";
	
	public String getLastICSUser() {
		return lastICSUser;
	}


	public void setLastICSUser(String lastICSUser) {
		this.lastICSUser = lastICSUser;
	}


	public String getLastICSPWD() {
		return lastICSPWD;
	}


	public void setLastICSPWD(String lastICSPWD) {
		this.lastICSPWD = lastICSPWD;
	}

	private File cfgFile = new File(System.getProperty("user.home")
			+"/.klaus/configuration.xml");
	
	private HashMap<String, String> persistence = new CFGHashMap();
	
	/**
	 * <p>Play sounds?</p>
	 * <p>Note this is not made persistent because this should only be set temporarily.
	 */
	private boolean playSounds = true;
	
	public Settings(){
		try {
			refreshConfig();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	
	@SuppressWarnings("unchecked")
	public void refreshConfig() throws JDOMException, IOException{
		System.out.println("Reading config, please wait.");
		SAXBuilder builder = new SAXBuilder();
        Document cfg;
        try{
            System.out.println("cfgFile is \""+cfgFile+"\" in Settings#refreshConfig()");
		    cfg = builder.build(cfgFile);
        }catch(Exception e){
            JOptionPane.showMessageDialog(null, format(Resources.getMessages().getString("invalid.config"), e.getMessage()));
            e.printStackTrace();
            //No config file, create it!
           /* if(cfgFile.exists()){
                cfgFile.renameTo(cfgFile.getName()+".backup");
                System.out.println("Dropped old (corrupt) config file with backup");
            }*/
            System.err.println("Config file does not exist, trying to create it in Settings#refreshConfig()");
            save();
            try{
                cfg = builder.build(cfgFile);
            }catch(IOException ex){
                throw new IOException("Failed to create config file in Settings#refreshConfig()!");
            }
        }
		Element root = cfg.getRootElement();
		
		List props = root.getChildren("property");

        for (Object prop1 : props) {
            Element property = (Element) prop1;
            if (property.getAttribute("name").getValue().equals("layout")) {
                layoutName = property.getAttributeValue("value");
            } else if (property.getAttributeValue("name").equals("username")) {
                username = property.getAttributeValue("value");
                System.out.println("Found username: "+username+" in Settings#refreshSettings()");
            } else if (property.getAttributeValue("name").equals("clockmode")) {
                clockMode = property.getAttributeValue("value");
            } else if (property.getAttributeValue("name").equals("icsuser")) {
                lastICSUser = property.getAttributeValue("value");
            } else if (property.getAttributeValue("name").equals("icspwd")) {
                lastICSPWD = property.getAttributeValue("value");
            }
        }
	
		//Persistence scheme is read
		for(Object obj: root.getChild("persistence").getChildren("property")){
			Element persistence = (Element)obj;
			
			this.persistence.put(persistence.getAttributeValue("name"), persistence.getAttributeValue("value"));
			
		}
		
		//get all available engines
		engines.clear();
		for(Object obj: root.getChildren("engine")){
			Element engine = (Element) obj;
			String engName = engine.getAttributeValue("id");
			System.out.println("\n\nFound engine: "+engName);
			String executable="";
			int protocol = 0;
			for(Object prop: engine.getChildren("property")){
				Element property = (Element) prop;
				if(property.getAttributeValue("name").equals("executable")){
					executable = property.getAttributeValue("value");
					System.out.println("Executable is "+executable);
				}else if(property.getAttributeValue("name").equals("protocol")){
					String prot = property.getAttributeValue("value");
					if(prot.equals("winboard")||prot.equals("xboard"))
						protocol = EngineData.ENGINE_PROTOCOL_WINBOARD_XBOARD;
					else if(prot.equals("uci"))
						protocol = EngineData.ENGINE_PROTOCOL_UCI;
					
					System.out.println("Protocol is "+prot);
				}
			}
			
			if(engine.getChild("default") != null){
				defaultEngine = engines.size();
				System.out.println("Found default engine: "+defaultEngine);
			}
			
			if(protocol != 0 && !executable.equals("")){
				System.out.println("Engine eingelesen.");
				EngineData data = new EngineData(protocol, executable, engName);
				engines.add(data);
			}
			else
				System.out.println("Enginedaten unvollständig, wird übersprungen!");
			
			System.out.println("\n\n");
			
		}
	}


	public String getLayoutName() {
		return layoutName;
	}


	public void setLayoutName(String layoutName) {
		this.layoutName = layoutName;
	}
	
	/**
	 * Get all available engines' data
	 * @return an array containing information about all available engines.
	 */
	public EngineData[] getEngines(){
		return engines.toArray(new EngineData[]{});
	}


	public String getUsername() {
		return username;
	}


	public void setUsername(String username) {
        System.out.println("Username set to "+username+" in Settings#setUsername(String)");
		this.username = username;
	}
	
	public void save(){
        System.out.println("Saving config in Settings.save()");
		String content = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n<configuration>\n"
		 + "\t<property name=\"layout\" value=\""+layoutName+"\"/>\n"
		 + "\t<property name=\"username\" value=\""+username+"\"/>\n"
		 + "\t<property name=\"clockmode\" value=\""+clockMode+"\"/>\n"
		 + "\t<property name=\"icsuser\" value=\""+lastICSUser+"\"/>\n"
		 + "\t<property name=\"icspwd\" value=\""+lastICSPWD+"\"/>\n\n"
		 + "\t<persistence>\n";
		
		for(String key: persistence.keySet().toArray(new String[]{})){
			content += "\t\t<property name=\""+key+"\" value=\""+persistence.get(key)+"\" />\n";
		}
		content += "\t</persistence>\n";
		
		int i = -1;
		for(EngineData data: engines){
			++i;
			content += "\t<engine id=\""+data.getName()+"\">\n";
            //Ths replace operation is necessary in order to handle "" in the executable's path. They have to be
            // escaped in order to prevent the config file getting corrupt and being dropped by refreshConfig().
			content += "\t\t<property name=\"executable\" value=\""+data.getExecutable().replace("\"", "&quot;")+"\"/>\n";
			content += "\t\t<property name=\"protocol\" value=\"";
			if(data.getEngineType() == EngineData.ENGINE_PROTOCOL_UCI)
				content += "uci";
			else
				content += "xboard";
			content += "\"/>\n";
			
			if(i == defaultEngine)
				content += "\t\t<default />\n";
			
			content += "\t</engine>\n\n";
		}
		
		content += "</configuration>";
		
		//If the cfg dir does not exist, create it
		System.out.println("Data dir is "+cfgFile.getParentFile().getAbsolutePath());
		if(!cfgFile.exists()) if(!cfgFile.getParentFile().mkdirs())
                System.err.println("Could not create cfg directory (Settings.save())");
		   /*
		ListFile.addToFile(FileSystemView.getFileSystemView().getHomeDirectory().getAbsolutePath()
				+"/.klaus/configuration.xml", content);            */
        try{
            PrintStream pw = new PrintStream(cfgFile);
            pw.println(content);
            pw.flush();
            pw.close();
            System.out.println("Saving settings successful in Settings#save()");
        }catch(Exception e){
            e.printStackTrace();
        }


	}


	public String getClockMode() {
		return clockMode;
	}
	
	public void setClockMode(String clockMode){
		this.clockMode = clockMode;
	}


	public void addEngine(EngineData data) {
		engines.add(data);
	}


	public void replaceEngine(EngineData old, EngineData newEng) {
		engines.set(engines.indexOf(old), newEng);
	}


	public void dropEngine(EngineData eng) {
		int index = engines.indexOf(eng);
		if(index < defaultEngine) --defaultEngine;
		else if(index == defaultEngine) defaultEngine = -1;
		engines.remove(eng);
	}


	public String getDefaultEngine() {
		if(defaultEngine >= 0)
			return engines.get(defaultEngine).getName();
		System.out.println("Settings#getDefaultEngine(): No defaultEngine set!");
		return Error.ERROR_0001_NO_DEFAULT_ENGINE;
	}


	public void setDefaultEngine(EngineData data) {
		defaultEngine = engines.indexOf(data);
	}


	public boolean isPlaySounds() {
		return playSounds;
	}


	public void setPlaySounds(boolean playSounds) {
		this.playSounds = playSounds;
	}
	
	public void setPersistence(String key, String value){
		persistence.put(key, value);
	}

    /**
     * Get a persistence setting's value
     * @param key the key
     * @param defValue a default value, this will be returned if the key was not set.
     * @return
     */
	public String getPersistence(String key, String defValue){
        if(persistence.containsKey(key))
		    return persistence.get(key);
        //This setting was never set before, save the default!
        setPersistence(key, defValue);
        return defValue;
	}
	
}
