/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.backend;

import klaus.chess.Move;
import klaus.chess.Position;
import klaus.chess.Result;
import klaus.chess.party.Party;

public interface ChessBackend {
	
	/**
	 * Set the backend active (let it accept new moves). If the backend is inactive,
	 * the frontend will not accept any UI moves as input.
	 *
	 */
	public void setActive();
	/**
	 * Set the backend inactive (see #setActive for more information)
	 *
	 */
	public void setInactive();
	
	/**
	 * Add a certain move to the backend. The backend has to react to this move.
	 * If the move is accepted, true is returned and the move can be displayed.
	 * If not, the interface should beep and/or show an error message.
	 * @param move the move that has been done by the user
	 * @return true if the move was accepted, else false
	 */
	public boolean addMove(Move move);
	/**
	 * Sets the backend position to the given position. If the user
	 * takes back some moves, the position has to be reset manually using
	 * this method.
	 * @param position the given position
	 */
	public void setPosition(Position position);
	
	public Position getPosition();
	
	/**
	 * Gives the remaining time for white [in milliseconds]
	 * @return 
	 */
	public long getWhiteTime();
	
	/**
	 * Gives the remaining time for black [im milliseconds]
	 * @return
	 */
	public long getBlackTime();
	
	public Party getParty();
	
	/**
	 * Unload the backend, e. g. close file/database/process connections
	 *
	 */
	public void unload();
	
	/**
	 * Add a move listener. Move listeners are fired everytime the backend
	 * does a move. Frontend moves are not handled.
	 * @param listener
	 */
	public void addMoveListener(MoveListener listener);
	/**
	 * Remove a move listener
	 * @param listener
	 */
	public void removeMoveListener(MoveListener listener);
	public void fireResultListeners(Result result);
	
	public void addResultListener(ResultListener listener);
	public void removeResultListener(ResultListener listener);
	public void fireMoveListeners(Move move);
	
	public boolean isGameRunning();
}
