/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.backend;

import klaus.chess.Move;
import klaus.chess.MoveControl;
import klaus.chess.Position;
import klaus.chess.Result;
import klaus.chess.party.Party;
import klaus.dialogs.NewGameModel;
import klaus.engine.Engine;
import klaus.sound.PlaySound;

public class OfflinePlayingBackend extends NullBackend implements ScrollingBackend {
	private class UpdateThread extends Thread {
		@Override
		public void run() {
			//Wait for start command
			while(!started){
				if(stopped)
					return;
				
				try{
					Thread.sleep(100);
				}catch(Exception e){}
			}
			while(!stopped){
				updateTimes();
				try{
					Thread.sleep(100);
				}catch(Exception e){}
			}
		}
	}
	
	private long lastUpdateTime = System.currentTimeMillis();
	
	protected Party party = new Party();
	
	//Flags for thread state
	private boolean started = false, stopped = false;
	
	private Engine white = null,
		black = null;
	
	private long whiteTime = 60000, blackTime = 60000;

    /**
     * Values for time on startup used for engine time management
     */
    private int startingTimeInMinutes = 0;

	private boolean gameRunning = true;
	
	/**
	 * The current ply number
	 */
	private int plyPointer = 0;
	
	public OfflinePlayingBackend(){
		new UpdateThread().start();
	}

	@Override
	public boolean addMove(Move move){
//		new Exception().printStackTrace();
		System.out.println("OfflinePlayingBackend#addMove(Move) called with move=\""+move.toXEngString()+"\"");
		if(!gameRunning) return false;
		
		if(active && MoveControl.accept(move, getPosition())){		
			party.addMove(move);
			System.out.println("OfflinePlayingBackend: Move added");
			PlaySound.playSoundclip("./sound/click.wav");

            if(getPosition().isCheckmate()){
				System.out.println("OfflinePlayingBackend: firing ResultListeners");
				if(getPosition().isWhitesTurn())
					fireResultListeners(Result.whiteCheckmated);
				else
					fireResultListeners(Result.blackCheckmated);
			}

			
			if(party.getPosition().isWhitesTurn() && white != null){
				++plyPointer;
				System.out.println("OfflinePlayingBackend: White engine starting...");
				white.addMove(move);
				white.startCalculation();
				if(black != null)
					black.cancel();
				setInactive();
			}if((!party.getPosition().isWhitesTurn())&& black != null){
				System.out.println("OfflinePlayingBackend: Black engine starting...");
				black.addMove(move);
				black.startCalculation();
				if(white != null)
					white.cancel();
				setInactive();
				//++plyPointer;
			}
			

			//Start the clock (if not done yet)
			++plyPointer;
			incremented = false;
			lastUpdateTime = System.currentTimeMillis();
			started = true;
			
			return true;
		}
		return false;
	}

	public void forceMove(){
		if(getPosition().isWhitesTurn() && white != null)
			white.forceMove();
		else if(black != null)
			black.forceMove();
	}
	
	public Engine getBlack() {
		return black;
	}
	
	@Override
	public long getBlackTime() {
		return blackTime;
	}

	@Override
	public Party getParty() {
		return party;
	}

	@Override
	public Position getPosition() {
		return party.getPosition(plyPointer);
	}

	public Engine getWhite() {
		return white;
	}

	@Override
	public long getWhiteTime() {
		return whiteTime;
	}
	
	public void goToFirst() {
		plyPointer = 0;
		fireMoveListeners(Move.NO_NEW_MOVE);
	}
	
	public void goToLast() {
		plyPointer = party.getPlyCount();
		fireMoveListeners(Move.NO_NEW_MOVE);
	}
	
	public void goToNext() {
		if(plyPointer < party.getPlyCount() - 1)
			++plyPointer;
		fireMoveListeners(Move.NO_NEW_MOVE);
	}
	
	public void goToPrevious() {
		if(plyPointer > 0)
			--plyPointer;
		fireMoveListeners(Move.NO_NEW_MOVE);
	}

	public boolean isEngineThinking(){
		if(active) return false;
		return true;
	}
	
	public void setBlackEngine(Engine e){
		System.out.println("OfflinePlayingBackend: Added black engine");
		if(black != null)
			black.removeAllMoveListeners();
		black = e;
        black.setTimeModel(startingTimeInMinutes, secondPeriodStart, increment/1000);
		black.addMoveListener(new MoveListener(){
			public void moveDone(Move move) {
				if(!gameRunning) return;
				if(white != null){
					white.addMove(move);
					white.startCalculation();
				}
				else
					setActive();
//				addMove(move);
				++plyPointer;
				fireMoveListeners(move);
			}
		});
	}        
	
	public void setWhiteEngine(Engine e){
		//To avoid funny effects, delete all listeners from the old engine...
		System.out.println("OfflinePlayingBackend: Added white engine");
		if(white != null)
			white.removeAllMoveListeners();
		white = e;
        white.setTimeModel(startingTimeInMinutes, secondPeriodStart, increment/1000);
		white.addMoveListener(new MoveListener(){
			public void moveDone(Move move) {
				if(!gameRunning) return;
				if(black!=null){
					black.addMove(move);
					black.startCalculation();
				}else
					setActive();
//				addMove(move);
				++plyPointer;
				fireMoveListeners(move);
			}
		});
		
	}
	
	public void startCalculation(){
		if(getPosition().isWhitesTurn() && white != null){
			white.startCalculation();
		}else if((!getPosition().isWhitesTurn()) && black != null){
			black.startCalculation();
		}
	}
	@Override
	public void unload(){
		if(white != null)
			white.unload();
		if(black != null)
			black.unload();
		
		stopped = true;
	}
	
	/**
	 * Update the clocks for both players. This method is called a) when the game is started and 
	 * b) by the UpdateThread ca. each 100 miliseconds.
	 */
	private synchronized void updateTimes(){
		long diff = System.currentTimeMillis() - lastUpdateTime;
		
		if(getParty().getPosition().isWhitesTurn())
			whiteTime -= diff;
		else
			blackTime -= diff;
		
		lastUpdateTime += diff;
		
		if(whiteTime < 0){
			whiteTime = 0;
			gameRunning = false;
			fireResultListeners(Result.whiteTimeout);
		}
		if(blackTime < 0){
			blackTime = 0;
			gameRunning = false;
			fireResultListeners(Result.blackTimeout);
		}
		
		if(!gameRunning) return;
		
		if(plyPointer == secondPeriodStart && !secondStarted){
			whiteTime += secondPeriod;
			blackTime += secondPeriod;
			secondStarted = true;
		}
		if(plyPointer == thirdPeriodStart && !thirdStarted){
			whiteTime += thirdPeriod;
			blackTime += thirdPeriod;
			thirdStarted = true;
		}
		if(incremented) return;
		if(!party.getPosition().isWhitesTurn())
			whiteTime += increment;
		else
			blackTime += increment;
		incremented = true;
	}
	
	private int secondPeriodStart = 0;
	private int secondPeriod = 0;
	private int thirdPeriodStart = 0;
	private int thirdPeriod = 0;
	private int increment = 0;
	private boolean secondStarted = false, thirdStarted = false, incremented = false;
	@Override
	public void fireResultListeners(Result result) {
		if(!gameRunning) return;
		gameRunning = false;
		super.fireResultListeners(result);
	}

	public void setGameModel(NewGameModel model) {
		System.out.println("Time model set!");
		blackTime = model.getFirstPeriod();
		whiteTime = model.getFirstPeriod();
		secondPeriodStart = model.getFirstPeriodDuration() * 2;
		secondPeriod = model.getSecondPeriod();
		thirdPeriodStart = (model.getFirstPeriodDuration()+model.getSecondPeriodDuration()) * 2;
		thirdPeriod = model.getThirdPeriod();
		increment = model.getIncrement();
		System.out.println("Time is "+blackTime+"; after "+secondPeriodStart+" + "+secondPeriod+"; after "+
				thirdPeriodStart+" + "+thirdPeriod+"; inc. "+increment);
        //Save starting time for engine init
        startingTimeInMinutes = (int)(whiteTime / 60000);
        // Set time for engines
        if(white != null){
            white.setTimeModel(startingTimeInMinutes, secondPeriodStart, increment/1000);
        }
        if(black != null){
            black.setTimeModel(startingTimeInMinutes, secondPeriodStart, increment/1000);
        }
	}
	
	@Override
	public boolean isGameRunning() {
		return gameRunning;
	}
	
}
