/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.chess;

import java.awt.*;

/**
 * Represents a half-move
 * @author Stephan Hillebrand
 * @since 2009-03-13
 */
public class Move {
	public static final Move NO_NEW_MOVE = new Move(-1,-1,-1,-1);
	
	/**
	 * The original X coordinate
	 */
	private int fromX = -1;
	/**
	 * The original Y coordinate
	 */
	private int fromY = -1;
	
	/**
	 * The new X coordinate
	 */
	private int toX = -1;
	/**
	 * The new Y coordinate
	 */
	private int toY = -1;
	
	private char promotion = ' ';
	
	private static final char[] lines = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'};
	private static final char[] rows = {'8', '7', '6', '5', '4', '3', '2', '1'};
	
	/**
	 * Creates a new move
	 * @param fromX the original location's X coordinate
	 * @param fromY the original location's Y coordinate
	 * @param toX the new X coordinate
	 * @param toY the new Y coordinate
	 */
	public Move(int fromX, int fromY, int toX, int toY){
		this.fromX = fromX;
		this.fromY = fromY;
		
		this.toX = toX;
		this.toY = toY;
	}
	
	/**
	 * Creates a new move with pawn promotion
	 * @param fromX the original location's X coordinate
	 * @param fromY the original location's Y coordinate
	 * @param toX the new X coordinate
	 * @param toY the new Y coordinate
	 * @param promotion the piece to which the pawn is promoted. ' ' will
	 * not do any promotion. Possibilities are QqRrNnBb or a blank. Upper /lower case
	 * is ignored.
	 */
	public Move(int fromX, int fromY, int toX, int toY, char promotion){
		this(fromX, fromY, toX, toY);
		setPromotion(promotion);
	}

	/**
	 * Creates a move from an engine's string
	 * @param engineMove the string sent by the engine. On WinBoard/XBoard and UCI this has the format
	 * 	e2e4 etc.
	 * @param position the old position, necessary e. g. for castling
	 */
	public Move(String engineMove, Position position) {
		engineMove = engineMove.trim();
//		System.out.println("\n\nMove#Move(String,Position):Parsing engine move "+engineMove);
		//Determine if this is a normal XBoard move like e2e4 or algebraic notation
		if(engineMove.matches("\\b[a-h][1-8][a-h][1-8][qnrbQNRB]?\\b")){
			//This is a normal move, maybe with promotion
			
			char ch1 = engineMove.charAt(0);
			char ch2 = engineMove.charAt(2);
			for(int i = 0; i < 8; i++){
				if(lines[i] == ch1)
					fromX = i;
				if(lines[i] == ch2)
					toX = i;
			}
			
			ch1 = engineMove.charAt(1);
			ch2 = engineMove.charAt(3);
			for(int i = 0; i < 8; i++){
				if(rows[i] == ch1)
					fromY = i;
				if(rows[i] == ch2)
					toY = i;
			}
			
			if(engineMove.length() == 5)
				promotion = engineMove.charAt(4);
		}
		
		else if(engineMove.matches("[KQRBNP]?.?[a-h][1-8][x\\-][a-h][1-8].*")){
			//Algebraic long notation
			//TODO Implement this
		}
		
		/*
		 * A move consists of a capital letter for a (k)ing, (q)ueen, (r)ook, (k)night
		 * or (b)ishop or nothing at all for a pawn, then zero or one letter to show where
		 * the piece was placed before if the move were arbitrary without this information
		 * or a pawn took a piece, zero or one "x" to show that a piece  was taken, the
		 * line and row of the piece and zero, one or more letters to show that this move
		 *  is check, checkmate, en passant or a pawn is promoted.
		 */ 
		else if(engineMove.matches(".*[a-h][1-8].*")){
			System.out.println("Move#Move(String,Position): "+engineMove+" is algebraic short notation!");
			//this is algebraic short notation!
			//get the piece..
			char piece = engineMove.charAt(0);
			//the first letter is not a piece -> it seems to be a pawn
			if(piece != 'K' && piece != 'Q' && piece != 'R' && piece != 'N' && piece != 'B')
				piece = 'P';
			System.out.println(piece);
			//Check who is moving and what is the piece's color
			if(!position.isWhitesTurn())
				piece = Character.toLowerCase(piece);
//			System.out.println("Move#Move(String,Position):Piece is "+piece);
			
			int lastDigit = -1;
			//Find last digit in the engine's move, this is the target square's digit!
			for(int i = engineMove.length() - 1; i > 0; --i){
				if(Character.isDigit(engineMove.charAt(i))){
					lastDigit = i;
					break;
				}
			}
//			System.out.println("Move#Move(String,Position):Last digit's position is "+lastDigit);
			
			if(lastDigit < 1){
//				System.out.println("Move#Move(String,Position):No target square digit found!");
				return;
			}
			char ch1 = engineMove.charAt(lastDigit);
			char ch2 = engineMove.charAt(lastDigit -1);
			for(int i = 0; i < 8; i++){
				if(rows[i] == ch1)
					toY = i;
				if(lines[i] == ch2)
					toX = i;
			}
//			System.out.println("Move#Move(String,Position):toX is "+toX+", toY is "+toY);
			
			fromX = -1;
			fromY = -1;
			
			boolean ambigous = false;
			//Get the starting position of the piece
			for(Point point: position.getPositionsOf(piece)){
				if(MoveControl.accept(new Move(
						(int)point.getX(),
						(int)point.getY(),
						getToX(),
						getToY())
				, position)){
					if(fromX < 0 || toX < 0){
						fromX = (int)point.getX();
						fromY = (int)point.getY();
//						System.out.println("Move#Move(String,Position):Found possible start position at "+fromX+"x"+fromY);
					}else{
//						System.out.println("Move#Move(String,Position):Move is ambigous, continue checking...");
						ambigous = true;
						
					}
				}		
			}
//			System.out.println("Move#Move(String,Position):fromX is "+fromX+", fromY is "+fromY);
			if(!ambigous)
				//Everything is done now, return!
				return;
			
			//continue the check as the move is ambigous
			char helper = ' ';
			
			//If the piece is a pawn, the helper is the first sign, else it's the second
			if(piece == 'P' || piece == 'p')
				helper = engineMove.charAt(0);
			else
				helper = engineMove.charAt(1);
			System.out.println("Move#Move(String,Position):Helper is "+helper);
			
			System.out.println("Helper ind is "+"abcdefgh12345678".indexOf(""+helper));
			if("abcdefgh12345678".indexOf(""+helper) < 0){
				System.out.println("Move "+engineMove+": Invalid helper!");
			}
			
			int rowSearch = "abcdefgh".indexOf(""+helper);
			int lineSearch = "87654321".indexOf(""+helper);
			System.out.println("Values are "+rowSearch+", "+lineSearch);
			for(Point point: position.getPositionsOf(piece)){
				if((rowSearch < 0 || (int)point.getY() == rowSearch)
					&&
					(lineSearch < 0 || (int)point.getX() == lineSearch)
					&&
					(Character.toUpperCase(position.getPieceAt((int)point.getX(), (int)point.getY()))
						== Character.toUpperCase(piece))
					&&
					MoveControl.accept(new Move((int)point.getX(), (int)point.getY(), toX, toY), position)
				){
					setFromX((int) getFromX());
					setFromY((int) getFromY());
					System.out.println("Match at "+fromX+"x"+fromY);
					return;
				}
			}
			System.out.println("Move#Move(String,Position): No matching piece found!");
		}
		
		else if(engineMove.trim().toLowerCase().equals("o-o") || engineMove.trim().equals("0-0")){
//			System.out.println("Move#Move(String,Position):Is short castling!");
			if(position.isWhitesTurn()){
				fromY = 7;
				toY = 7;
			}
			else{
				fromY = 0;
				toY = 0;
			}
			
			fromX = 4;
			toX = 6;
				
		}
		else if(engineMove.trim().toLowerCase().equals("o-o-o") || engineMove.trim().equals("0-0-0")){
//			System.out.println("Move#Move(String,Position):Is long castling!");
			if(position.isWhitesTurn()){
				fromY = 7;
				toY = 7;
			}
			else{
				fromY = 0;
				toY = 0;
			}
			fromX = 4;
			toX = 2;
		}else{
//			System.out.println("Move#Move(String,Position):Move format was not determined correctly.");
			return;
		}
		
	}

	public int getFromX() {
		return fromX;
	}

	public int getFromY() {
		return fromY;
	}

	public int getToX() {
		return toX;
	}

	public int getToY() {
		return toY;
	}

	public void setFromX(int fromX) {
		this.fromX = fromX;
	}

	public void setFromY(int fromY) {
		this.fromY = fromY;
	}

	public void setToX(int toX) {
		this.toX = toX;
	}

	public void setToY(int toY) {
		this.toY = toY;
	}

	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + fromX;
		result = PRIME * result + fromY;
		result = PRIME * result + toX;
		result = PRIME * result + toY;
		return result;
	}
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final Move other = (Move) obj;
		if (fromX != other.fromX)
			return false;
		if (fromY != other.fromY)
			return false;
		if (toX != other.toX)
			return false;
		if (toY != other.toY)
			return false;
		return true;
	}
	
	/**
	 * This applies this move to a position and returns a new position with the
	 * given move done. The original position is not touched.
	 * This method can also (but does not necessarily) check if the move is legal.
	 * If this will be done can be set using the check variable.
	 * 
	 * Note that the move is not given as an argument! This is an instance method of
	 * Move.
	 *
	 *
	 * @param pos the original position
	 * 
	 * @return a clone of the original position with the move done.
	 */
	public Position applyMoveTo(Position pos){
		Position working = (Position) pos.clone();

		//Is there a piece at all??? Of course it is nonsense to move a empty square!
		piece = pos.getPieceAt(getFromX(), getFromY());
		if(piece == ' '){
			return working;
		}

		if(piece == 'Q')
			System.out.println("");
			
		char taken = pos.getPieceAt(getToX(), getToY());
		if(taken != ' '){
			take = true;
		}
		
		//If the move does not do anything, that is both squares are the same,
		//nothing is done.
		if(getFromX() == getToX() && getFromY() == getToY()) return working;
		
		working.setEnPassantPossibility(-1);
		
		//Move the piece to the new square
		working.setPieceAt(getToX(), getToY(), piece);
		//make the old square empty
		working.setPieceAt(getFromX(), getFromY(), ' ');
		
		if(piece == 'k'){
			if(getFromX() == 4 && getToX() == 2 && getToY() == 0 && getFromY() == 0){
				//Black castles on the queen's side
				//place the rook on (3|0) and clear (0|0)
				working.setPieceAt(0, 0, ' ');
				working.setPieceAt(3, 0, 'r');
//				System.out.println("Move:                  Is long castling (B)");
			}else if(getFromX() == 4 && getToX() == 6 && getFromY() == 0 && getToY() == 0){
				working.setPieceAt(7, 0, ' ');
				working.setPieceAt(5, 0, 'r');
//				System.out.println("Move:                  Is short castling (B)");
			}
			
			working.setLongCastlingBlack(false);
			working.setShortCastlingBlack(false);
		}else if(piece == 'K'){
			if(getFromX() == 4 && getToX() == 2 && getToY() == 7 && getFromY() == 7){
				//White castles on the queen's side
				//place the rook on (3|7) and clear (0|7)
				working.setPieceAt(0, 7, ' ');
				working.setPieceAt(3, 7, 'R');
//				System.out.println("Move:                  Is long castling (W)");
			}else if(getFromX() == 4 && getToX() == 6 && getFromY() == 7 && getToY() == 7){
				working.setPieceAt(7, 7, ' ');
				working.setPieceAt(5, 7, 'R');
//				System.out.println("Move:                  Is short castling (W)");
			}
			
			working.setLongCastlingWhite(false);
			working.setShortCastlingWhite(false);
		}else if(piece == 'P' || piece == 'p'){
			//Double step -> set en passant active!
			
			if(Math.abs(getFromY()-getToY()) == 2){
				working.setEnPassantPossibility(getToX());
//				System.out.println("Move:                  Is double step");
			}
			else if(piece == 'p' && getFromY() == 4 && getToY() == 5 && Math.abs(getFromX()-getToX()) == 1
					&& taken == ' '){
				working.setPieceAt(getToX(), 4, ' ');
				take = true;
				ep = true;
//				System.out.println("Move:                  Is en passant (B)");
			}
			else if(piece == 'P' && getFromY() == 3 && getToY() == 2 && Math.abs(getFromX()-getToX()) == 1
					&& taken == ' '){
				working.setPieceAt(getToX(), 3, ' ');
				take = true;
				ep = true;
//				System.out.println("Move:                  Is en passant (W)");
			}
			
			//Promoting!
			if((piece == 'P' &&getToY() == 0) ||(piece == 'p' && getToY() == 7)){
				working.setPieceAt(getToX(), getToY(), Character.isUpperCase(piece) ? 
						Character.toUpperCase(getPromotion()) : Character.toLowerCase(getPromotion()));
				
//				System.out.println("Move:                  Is promoting");
			}


            // If a rook moved update castling rights
            if(getFromX() == 0 && getFromY() == 0) pos.setLongCastlingBlack(false);
            if(getFromX() == 0 && getFromY() == 8) pos.setShortCastlingBlack(false);
		    if(getFromX() == 8 && getFromY() == 0) pos.setLongCastlingWhite(false);
            if(getFromX() == 8 && getFromY() == 8) pos.setShortCastlingWhite(false);
        }
		
		//Change whitesTurn
		working.setWhitesTurn(!working.isWhitesTurn());
		
		/*//Some things for notation
		 *Unfortunately, this causes an endless recursion.
		if(working.isCheckmate()) checkmate = true;
		else if(working.isCheck()) check = true;
		else if(working.isStalemate()) stalemate = true;*/
		newPosition = (Position) working.clone();
		return working;
	}
	
	private Position newPosition = null;

	private char piece = ' ';
	private boolean check = false;
	private boolean checkmate = false;
	private boolean stalemate = false;
	private boolean take = false;
	private boolean ep = false;
	private boolean longCastling = false;
	private boolean shortCastling = false;
	
	public String getNotation(@SuppressWarnings("unused")
	Position oldPosition) { 
		if(newPosition != null){
			if(newPosition.isCheckmate()) checkmate = true;
			else if(newPosition.isCheck()) check = true;
			else if(newPosition.isStalemate()) stalemate = true;
		}
		
		if(longCastling) return "O-O-O";
		if(shortCastling) return "O-O-O";
		
		
		//Give a notation for this move
		String not = "";
		if(Character.toUpperCase(piece) != 'P')
			not += toLocalizedPiece(Character.toUpperCase(piece))+"";
		
		not += ""+lines[getFromX()]+""+rows[getFromY()]+
		 (take ? "x" : "-")
		+lines[getToX()]+""+rows[getToY()];
		
		if(ep) not+= " e. p.";
		if(check) not += "+";
		if(checkmate) not += "#";
		if(stalemate) not += "\n1/2:1/2";
		return not.trim();
		
	}

	private char toLocalizedPiece(char c) {
		if(c == 'K' || c == 'k') return 'K';
		if(c == 'Q' || c == 'q') return 'D';
		if(c == 'R' || c == 'r') return 'T';
		if(c == 'N' || c == 'n') return 'S';
		if(c == 'B' || c == 'b') return 'L';
		return ' ';
	}

	/**
	 * Generate a description string for WinBoard/XBoard connections
	 * @return a String like e2e4 for communication with WinBoard / XBoard engines
	 */
	public String toXEngString() {
        try{
		    return lines[getFromX()]+""+rows[getFromY()]+""+lines[getToX()]+""+rows[getToY()];
        }catch(ArrayIndexOutOfBoundsException aioobe){
            return "Can't create move with coords "+getFromX()+"|"+getToX()+" - "+getFromX()+"|"+getToX();
        }
	}

	public char getPromotion() {
		return promotion;
	}

	public void setPromotion(char promotion) {
		this.promotion = promotion;
	}

}
