/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.chess;

import java.awt.*;


/**
 * This class is used to check if a move is legal, if a certain
 * square is controlled by white or black, if a king is in check etc.
 * 
 * @author Stephan Hillebrand
 * @since 2009-03-13
 */
public class MoveControl {
    /**
     * Check if a move is legal in a given position
     * @param move the move to be checked
     * @param position the position in which the move should be done
     * @return true if the move was accepted, false if it was rejected
     */
	public static boolean accept(Move move, Position position) {
		/*
		 * Check if the player whose turn it is did move. Otherwise it is illegal.
		 */
		if(position.isWhitesTurn() != Character.isUpperCase(position.getPieceAt(move.getFromX(), move.getFromY()))) 
			return false;
		//You can't take your own pieces! What did you expect? So check if a piece is
		//taken and, if yes, if it belongs the moving player.
		if(position.getPieceAt(move.getToX(), move.getToY()) != ' ' &&
				(Character.isUpperCase(position.getPieceAt(move.getToX(), move.getToY())) == position.isWhitesTurn())
		)
			return false;
		
		//check if the move's direction is rule-conform 
		if(!canDoMove(move, position))
			return false;
		
		//Check if the king is in check after this move
		Position after = move.applyMoveTo(position);
		Point king = findKing(after, position.isWhitesTurn());
		if(squareInCheck(after,(int)king.getX(), (int)king.getY(), position.isWhitesTurn()))
			return false;
		return true;
	}

    /**
     * Checks whether a move is technically possible (meaning regarding the pieces positions, if the way is free etc,
     * <b>ignoring if pieces are in check etc.</b>. For checking if a move is completely legal, use accept() instead.
     * @param move
     * @param position
     * @return
     */
	private static boolean canDoMove(Move move,Position position) {
		char piece = position.getPieceAt(move.getFromX(), move.getFromY());
		
		/*
		 * Check if this is a king's move and the king moved correctly (max. 1 square difference
		 * in X and Y direction
		 */
		if(piece == 'k' || piece == 'K'){
			if( Math.abs(move.getFromX() - move.getToX()) > 1 || Math.abs(move.getFromY() - move.getToY()) > 1 ){
				//This does not appear to be a normal move!
				//Maybe it is a castling????
				
				if(piece == 'k'){
					//Black castlings
					
					if(!(
							
							//Short castling, do not check the target square! This is done in accept()!
							(position.isShortCastlingBlack() && move.getFromX() == 4 && move.getToX() == 6
									&&move.getFromY() == 0 &&move.getToY() == 0
									&& (!squareInCheck(position,4, 0, false)) && (!squareInCheck(position,5, 0, false)))
							||
							//Long castling
							(position.isLongCastlingBlack() && move.getFromX() == 4 && move.getToX() == 2
									&&move.getFromY()  == 0 && move.getToY() == 0
									&& (!squareInCheck(position,4, 0, false)) && (!squareInCheck(position,3, 0, false)))
					
					)){
						
						//it is NOT a castling! As it is not a normal move either, it appears to be illegal.
						return false;
					}
				}else{
					//the same as above for white...
					
					if(!(
							//Short castling
							(position.isShortCastlingWhite() && move.getFromX() == 4 && move.getToX() == 6
									&&move.getFromY() == 7 &&move.getToY() == 7
									&& (!squareInCheck(position,4, 7, true)) && (!squareInCheck(position,5, 7, true)))
							||
							//Long castling
							(position.isLongCastlingWhite() && move.getFromX() == 4 && move.getToX() == 2
									&&move.getFromY()  == 7 && move.getToY() == 7
									&& (!squareInCheck(position,4, 7, true)) && (!squareInCheck(position,3, 7, true)))
					
					)){
						
						//it is NOT a castling! As it is not a normal move either, it appears to be illegal.
						return false;
					}
				
				}
				
			}
			//It is a legal normal move, so do nothing in order to let the program do the remaining
			//checks at the end of this function...
		}
		
		else if(piece == 'q' || piece == 'Q'){
			/*
			 * Queens are a bit more difficult to handle as they might look as they can move in two
			 * different directions. The following is a combination of the rooks' and bishops' codes.
			 */
			
			//First check horizontal and vertical moves
			if(move.getFromX() == move.getToX() ^ move.getFromY() == move.getToY()){
				//It is a correct move! Now check if no pieces are in the way...
				if(move.getFromX() == move.getToX()){
					//This is a vertical move
					//Get the direction
					int direction = 1;
					if (move.getToY() < move.getFromY())
						//Moving up => direction = -1
						direction = -1;
					
					for(int i = move.getFromY() + direction; i != move.getToY(); i += direction){
						if(position.getPieceAt(move.getFromX(), i) != ' ')
							//The square is not free! Return false!
							return false;
					}
				}else{
					//this is a horizontal move!
					//get the direction again
					int direction = 1;
					if(move.getFromX() > move.getToX())
						direction = -1;
					
					for(int i = move.getFromX() + direction; i != move.getToX(); i+=direction){
						if(position.getPieceAt(i, move.getFromY()) != ' ')
							return false;
					}
				}
				
				//Now check diagonal moves...
			}else if(Math.abs(move.getFromX()-move.getToX()) == Math.abs(move.getFromY() - move.getToY())){
				//check the directions
				int directionX = 1;
				if(move.getToX() < move.getFromX())
					directionX = -1;
				
				int directionY = 1;
				if(move.getToY() < move.getFromY())
					directionY = -1;
				
				// Check if everything is free
				int y = move.getFromY();
				for(int x = move.getFromX() + directionX; x != move.getToX(); x += directionX){
					y += directionY;
					
					if(position.getPieceAt(x, y) != ' ')
						//A piece is in the way -> move illegal!
						return false;
					
				}
				
			}
			else{
				return false;
			}
			
			
		}
		
		// Check rook moves
		else if(piece == 'r' || piece == 'R'){
			//Check f it is a correct move
			
			if(move.getFromX() == move.getToX() ^ move.getFromY() == move.getToY()){
				//It is a correct move! Now check if no pieces are in the way...
				if(move.getFromX() == move.getToX()){
					//This is a vertical move
					//Get the direction
					int direction = 1;
					if (move.getToY() < move.getFromY())
						//Moving up => direction = -1
						direction = -1;
					
					for(int i = move.getFromY() + direction; i != move.getToY(); i += direction){
						if(position.getPieceAt(move.getFromX(), i) != ' ')
							//The square is not free! Return false!
							return false;
					}
				}else{
					//this is a horizontal move!
					//get the direction again
					int direction = 1;
					if(move.getFromX() > move.getToX())
						direction = -1;
					
					for(int i = move.getFromX() + direction; i != move.getToX(); i+=direction){
						if(position.getPieceAt(i, move.getFromY()) != ' ')
							return false;
					}
				}
			}
			else
				return false;
			
		}
		/*
		 * Bishops
		 */
		else if((piece == 'B') || (piece == 'b')){
			
			//Check if deltaX and deltaY are equal
			if(Math.abs(move.getFromX()-move.getToX()) == Math.abs(move.getFromY() - move.getToY())){
				//check the directions
				int directionX = 1;
				if(move.getToX() < move.getFromX())
					directionX = -1;
				
				int directionY = 1;
				if(move.getToY() < move.getFromY())
					directionY = -1;
				
				// Check if everything is free
				int y = move.getFromY();
				for(int x = move.getFromX() + directionX; x != move.getToX(); x += directionX){
					y += directionY;
					
					if(position.getPieceAt(x, y) != ' ')
						//A piece is in the way -> move illegal!
						return false;
					
				}
				
			}
			else{
				return false;
			}
		}

		/*
		 * Knights are quite simple to implement as they can jump over any
		 * other pieces. So no check has to be done for the crossed squares.
		 */
		else if((piece == 'n')||piece == 'N'){
			if(!(			
					(Math.abs(move.getToX()-move.getFromX())==2 && Math.abs(move.getToY()-move.getFromY())==1)
					||
					(Math.abs(move.getToX()-move.getFromX())==1 && Math.abs(move.getToY()-move.getFromY())==2)
					//looking nasty, doesn't it?
			))
				//Not a normal knight move => return false!
				return false;
					
		}
		
		/*
		 * This is a  pawn->so there's a lot of work to be done: e. p., taking, double step etc.
		 */
		else if((piece == 'p')||(piece == 'P')){
			//Check if it's a white or a black pawn
			//-1 means the piece oves up (white). +1 is the opposite
			int direction = -1;
			if(piece == 'p')
				direction = 1;
			
			//Check if it's a normal move
			if(move.getToY() - move.getFromY() != direction || move.getToX() != move.getFromX()){
				//This is not a normal move!
				//Check if a piece is taken...
				if(Math.abs(move.getToX()-move.getFromX()) == 1){
					//A piece is taken
					if(!(
							
							//Normal taking
							(move.getToY() - move.getFromY() == direction && 
									position.getPieceAt(move.getToX(), move.getToY()) != ' ')
									
							||
							//en passant for white
							(piece == 'P' && position.getEnPassantPossibility() == move.getToX()
									&& move.getFromY() == 3 && move.getToY() == 2)
							||
							//en passant for black
							(piece == 'p' && position.getEnPassantPossibility() == move.getToX()
									&& move.getFromY() == 4 && move.getToY() == 5)
					))
						//Illegal taking!
						return false;
					//legal taking
					
				}else
					//maybe it's a double step?
					if(!(
							//white
						(piece == 'P' && move.getFromY() == 6 && move.getToY() == 4 && 
								move.getFromX() == move.getToX() && position.getPieceAt(move.getFromX(), 4) == ' '&&
								position.getPieceAt(move.getFromX(), 5) == ' ')
						||
						//black
						(piece == 'p' && move.getFromY() == 1 && move.getToY() == 3 &&
								move.getFromX() == move.getToX()&& position.getPieceAt(move.getFromX(),2) == ' '&&
								position.getPieceAt(move.getFromX(), 3) == ' ')
					))
						//It is not a correct double step!
						return false;
			}else{
				//It is a normal move! -> need to check if the target field is empty
				if(position.getPieceAt(move.getToX(), move.getToY()) != ' ')
					return false;
			}
			
		}
		else{
			//The piece was not identified! This is a quite weird error and should not
			//be accepted.
			return false;
		}
		
		
		//If no error was found, return true!
		return true;
	}

    /**
     * Get the king's current position
     * @param position a position to search
     * @param white true if the white king is to be searched, false if we need the black king's position
     * @return
     */
	public static Point findKing(Position position, boolean white){
		char piece = white ? 'K' : 'k';
		for(int x = 0; x < 8; x++){
			for(int y = 0; y < 8; y++){
				if (position.getPieceAt(x, y) == piece)
					return new Point(x, y);
			}
		}
		
		//This is not supposed to happen as the kings are always on board!
		System.out.println("MoveControl:   [ERROR]:No king found, dumping position...");
		position.dump(System.out);
		return new Point(-1,-1);
//		System.out.println("MoveControl:   [ERROR]:Illegal position: the king is missing!");
	}
	
	/**
	 * Checks if a certain square is in check
     * @param pos the position 
	 * @param searchx the square that has to be checked (X coordinate)
     * @param searchy the square that has to be checked (Y coordinate)
	 * @param white who shall be in check?
	 * @return true if the king is in check, else false
	 */
	public static boolean squareInCheck(Position pos, int searchx, int searchy, boolean white){
		
		for(int px = 0; px < 8; px++){
			for(int py = 0; py < 8; py++){
				if(px == searchx && py == searchy) continue;
				
				if((py != searchy || px != searchx) && pos.getPieceAt(px, py) != ' ' 
					&& Character.isUpperCase(pos.getPieceAt(px, py))!= white
					&&canDoMove(new Move(px, py, searchx, searchy), pos)){
					//System.out.println("Break at "+px+"x"+py+": Move "+new Move(px, py, searchx, searchy).getNotation());
					return true;
				}
			}
		}
		
		
		return false;
	}
	
}
