/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.chess;

import java.awt.*;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Vector;

/**
 * This class stores a position.
 * @author Stephan Hillebrand
 * @since 2009-03-11
 */
public class Position {

	/**
	 * Is it white's turn (is this a max position) ???
	 */
	private boolean whitesTurn = true;
	
	/**
	 * Number of half-moves since the last time a pawn moved or a piece
	 * was taken. If this number reaches 100 (that is, 50 moves) the game is
	 * drawn.
	 */
	private int noOfDrawMoves = 0;

    /**
     * Number of moves (not half-moves, notice the dirfference to noOfDrawMoves!!!)
     */
    private int noOfMove = 1;
	
	/**
	 * This variable stores the posiibilities to take en passant (e. p.).
	 * If the last move was a pawn's double step, this is set to the
	 * line number of this pawn (0 for A, 7 for H). If not, -1 will
	 * signal that there is no posiibility to take en passant.
	 * After each half-move this has to be reset.
	 */
	private int enPassantPossibility = -1;
	
	private boolean longCastlingWhite = true,
		longCastlingBlack = true,
		shortCastlingWhite = true,
		shortCastlingBlack = true;
	/**
	 * <p>This matrix is used to store the position. Capital letters are white
	 * pieces, other characters are black. The blanks are empty fields.</p>
	 * 
	 * <p>Each new position is initialized with the default position. I don't
	 * know if this is that good, but it makes things more easy.</p> 
	 */
	private char[][] pieces = {
			{'r', 'n', 'b', 'q', 'k', 'b', 'n', 'r'},
			{'p', 'p', 'p', 'p', 'p', 'p', 'p', 'p'},
			{' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
			{' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
			{' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
			{' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
			{'P', 'P', 'P', 'P', 'P', 'P', 'P', 'P'},
			{'R', 'N', 'B', 'Q', 'K', 'B', 'N', 'R'}
		};
	
	
	/**
	 * <p>This creates a new position from a FEN string. </p>.
	 * <p>The most simple way to create a Position is giving a FEN String.
	 * This might look like this:
	 * 
	 * <code>rbnqknbr/pppppppp/8/8/8/8/PPPPPPPP/RBNQKNBR w - - 0 1</code>
	 * 
	 * where the rbn... sequence is the board, w means it's white's move (b would
	 * mean black), - is the possibility to castle (in this example nobody can castle,
	 * Kq would mean white can do a short castling and black can do a long castling),
	 * the second - means nobody can take e. p. (else e. g. 4 for e. p. on the E line).
	 * 0 is the number of half-moves for the 50 moves rule and 1 is the number of this move (NOT
	 * half-move but move, really!)</p>
	 * 
	 * @param fen the FEN string used (Forsyth-Edwards notation, see Wikipedia for details)
	 */
	@SuppressWarnings("unused")
	public Position(String fen){
		//Ignore multiple white-spaces
		String[] args = fen.split(" +");
		
		String[] rows = args[0].split("/");
		if((rows.length != 8) || (args.length != 6)){
			System.out.println("Illegaler FEN-String!");
			return;
		}
		for(int i = 0; i < 8; ++i){
			// Pointer to a field
            System.out.println(i);
			int fieldPointer = 0;
			for(int pos = 0; pos < rows[i].length(); pos++){
                    char at = rows[i].charAt(pos);
                    if(Character.isDigit(at)){
                        for(int count = 0; count < Integer.parseInt(at+""); ++count){
                            setPieceAt(fieldPointer++, i, ' ');
                        }
                    }
                    else{
                        setPieceAt(fieldPointer++, i, at);
                    }
			}
		}

        // Parse whose turn it is
        if(args[1].equals("w")) whitesTurn = true;
        else whitesTurn = false;

        //Castlings
        if(!args[2].contains("K")) shortCastlingWhite = false;
        if(!args[2].contains("Q")) longCastlingWhite = false;
        if(!args[2].contains("k")) shortCastlingBlack = false;
        if(!args[2].contains("q")) longCastlingBlack = false;

        if(!args[3].equals("-")) enPassantPossibility = Integer.parseInt(""+args[3].charAt(0));

        noOfDrawMoves = Integer.parseInt(args[4]);
        setNoOfMove(Integer.parseInt(args[5]));

	}
	
	/**
	 * Creates a default position
	 */
	public Position(){
		
	}
	
	public void setPieceAt(int line, int row, char piece){
		pieces[row][line] = piece;
	}
	
	public char getPieceAt(int line, int row){
		return pieces[row][line];
	}

	public int getEnPassantPossibility() {
		return enPassantPossibility;
	}

	public int getNoOfDrawMoves() {
		return noOfDrawMoves;
	}

	public boolean isWhitesTurn() {
		return whitesTurn;
	}

	public void setEnPassantPossibility(int enPassantPossibility) {
		this.enPassantPossibility = enPassantPossibility;
	}

	public void setNoOfDrawMoves(int noOfDrawMoves) {
		this.noOfDrawMoves = noOfDrawMoves;
	}

	public void setWhitesTurn(boolean whitesTurn) {
		this.whitesTurn = whitesTurn;
	}
	
	
	@Override
	public Object clone(){
		Position clone = new Position();
		clone.setWhitesTurn(whitesTurn);
		clone.setEnPassantPossibility(enPassantPossibility);
		clone.setNoOfDrawMoves(noOfDrawMoves);
		clone.pieces = new char[8][8];
		for(int i = 0; i<8;i++){
			for(int j = 0; j<8; j++){
				clone.pieces[i][j] = pieces[i][j];
			}
		}
		
		clone.setShortCastlingBlack(shortCastlingBlack);
		clone.setShortCastlingWhite(shortCastlingWhite);
		clone.setLongCastlingBlack(longCastlingBlack);
		clone.setLongCastlingWhite(longCastlingWhite);
		
		return clone;
	}

	public boolean isLongCastlingBlack() {
		return longCastlingBlack;
	}

	public boolean isLongCastlingWhite() {
		return longCastlingWhite;
	}

	public boolean isShortCastlingBlack() {
		return shortCastlingBlack;
	}

	public boolean isShortCastlingWhite() {
		return shortCastlingWhite;
	}

	public void setLongCastlingBlack(boolean longCastlingBlack) {
		this.longCastlingBlack = longCastlingBlack;
	}

	public void setLongCastlingWhite(boolean longCastlingWhite) {
		this.longCastlingWhite = longCastlingWhite;
	}

	public void setShortCastlingBlack(boolean shortCastlingBlack) {
		this.shortCastlingBlack = shortCastlingBlack;
	}

	public void setShortCastlingWhite(boolean shortCastlingWhite) {
		this.shortCastlingWhite = shortCastlingWhite;
	}

	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + enPassantPossibility;
		result = PRIME * result + (longCastlingBlack ? 1231 : 1237);
		result = PRIME * result + (longCastlingWhite ? 1231 : 1237);
		result = PRIME * result + noOfDrawMoves;
		result = PRIME * result + Arrays.hashCode(pieces);
		result = PRIME * result + (shortCastlingBlack ? 1231 : 1237);
		result = PRIME * result + (shortCastlingWhite ? 1231 : 1237);
		result = PRIME * result + (whitesTurn ? 1231 : 1237);
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final Position other = (Position) obj;
		if (enPassantPossibility != other.enPassantPossibility)
			return false;
		if (longCastlingBlack != other.longCastlingBlack)
			return false;
		if (longCastlingWhite != other.longCastlingWhite)
			return false;
		if (noOfDrawMoves != other.noOfDrawMoves)
			return false;
		if (!Arrays.equals(pieces, other.pieces))
			return false;
		if (shortCastlingBlack != other.shortCastlingBlack)
			return false;
		if (shortCastlingWhite != other.shortCastlingWhite)
			return false;
		if (whitesTurn != other.whitesTurn)
			return false;
		return true;
	}
	
	/**
	 * check if the player moving next has any possible moves
	 * @return
	 */
	public boolean hasLegalMoves(){
		//Iterate over all pieces of the player
//		System.out.println("\nMove control:          Checking legal moves");
//		boolean result = false;
		for(int x1 = 0; x1 < 8; x1++){
			for(int y1 = 0; y1 < 8; y1++){
				if(Character.isUpperCase(getPieceAt(x1, y1)) == whitesTurn){
					//Iterate over all squares
					for(int x2 = 0; x2 < 8; x2++){
						for(int y2 = 0; y2 < 8; y2++){
							if(MoveControl.accept(new Move(x1, y1, x2, y2), (Position)this.clone())){
//								System.out.println("Move control:          Legal move: "+new Move(x1, y1, x2, y2).getNotation());
								return true;
							}
						}
					}

				}
			}
		}
		//System.out.println("Move control:          No legal move!");
		return false;
	}
	
	public boolean isCheck(){
		Point point = MoveControl.findKing(this, whitesTurn);
		if(MoveControl.squareInCheck(this, (int)point.getX(), (int)point.getY(), isWhitesTurn()))
			return true;
		return false;
	}
	
	public boolean isCheckmate(){
		if(((Position)clone()).isCheck()&& !((Position)clone()).hasLegalMoves())
			return true;
		return false;
	}
	
	public boolean isStalemate(){
		if(!((Position)clone()).isCheck() && !((Position)clone()).hasLegalMoves())
			return true;
		return false;
	}
	
	
	public void dump(PrintStream ps){
		ps.println("\n\nPosition dump: ");
		for(int i = 0; i < 8; i++){
			for(int j = 0; j < 8; j++){
				ps.print(getPieceAt(i, j));
			}
			ps.println();
		}
		ps.println("Moving player: "+(isWhitesTurn() ? "white": "black"));
		ps.println("\n\n\n");
	}
	
	
	public Point[] getPositionsOf(char piece){
		Vector<Point> points = new Vector<Point>();
		for(int x = 0; x < 8; ++x){
			for(int y = 0; y < 8; ++y){
				if(getPieceAt(x, y) == piece){
					
//					System.out.println("Position#getpointsOf(char): Found at "+x+"x"+y);
					points.add(new Point(x,y));
					
				}
			}
		}
		return points.toArray(new Point[]{});
	}

    /**
     * Generate a FEN (Forsyth-Edwards notation) string from this position.
     * @return a FEN string for this position
     */
    public String toFEN(){
        String res = "";

        //Resolve position
        for(int row = 0; row < 8; ++row){
            int count = 0;
            for(int line = 0; line < 8; ++line){
                if(getPieceAt(line, row) == ' ') ++count;

                else{
                    if(count != 0) res += ""+count;
                    count = 0;

                    res += getPieceAt(line, row);
                }

                if(line == 7){
                    if(count != 0) res += ""+count;
                    count = 0;

                    if(row != 7) res += "/";
                }
            }
        }
        //Add information about black and white

        res += isWhitesTurn() ? " w " : " b ";

        //Castling!
        boolean anyCastling = false;
        if(isShortCastlingWhite()){
            res += "K";
            anyCastling = true;
        }
        if(isLongCastlingWhite()){
            res += "Q";
            anyCastling = true;
        }
        if(isShortCastlingBlack()){
            res += "k";
            anyCastling = true;
        }
        if(isShortCastlingWhite()){
            res += "q";
            anyCastling = true;
        }

        //if no castlings were possible, add -
        if(!anyCastling) res += "-";

        res += " ";

        if(getEnPassantPossibility() < 0) res += "-";
        else res += new char[]{'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'}[getEnPassantPossibility()]
                + (isWhitesTurn() ? "6": "3");

        res += " " + getNoOfDrawMoves() + " " + getNoOfMove();

        return res;
    }

    public int getNoOfMove() {
        return noOfMove;
    }

    public void setNoOfMove(int noOfMove) {
        this.noOfMove = noOfMove;
    }



}
