/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.chess.party;

import klaus.Resources;
import klaus.chess.Move;
import klaus.chess.Position;
import myutils.ListFile;

import java.io.File;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;

/**
 * This represents a party.
 * @author Stephan Hillebrand
 * @version 2 - using a tree instead of a list
 */
public class Party {

	private String result = "";
	private String date = "", site="", tournament="", ECO = "";
	private String round = "";
	/**
	 * Name of the white/black player
	 */
	private String whitePlayer = "", blackPlayer = "";
	private String startingComment = "";
	
	private ArrayList<String> additionalMetaData = new ArrayList<String>();
	
	/**
	 * The party root node
	 */
	private PartyNode root = new PartyNode();
	
	/**
	 * The node before lastNode, that is the node where variations are added
	 */
	private PartyNode varNode = null;
	
	/**
	 * A reference to the last node of PV
	 */
	private PartyNode lastNode = root;
	/**
	 * The number of plys in the pv
	 */
	private int plyCount = 0;
	
	/**
	 * add a new move at the end of the PV
	 * @param m the move to add
	 */
	public void addMove(Move m){
		if(m.getFromX() < 0 || m.getToX() <0 || m.getFromX() > 7 || m.getToX() > 7
				|| m.getToY() < 0 || m.getToY() > 7 || m.getFromY() < 0 || m.getFromY() > 7)
			return;
		
		PartyNode next = new PartyNode(m);
		
		varNode = lastNode;
		lastNode.addPV(next);
		lastNode = next;
		plyCount++;
	}
	
	/**
	 * Adds a variation after the previous pv move
	 * @param variation the variation to add
	 */
	public void addVariation(PartyNode variation) {
		varNode.addVariant(variation);
	}
	
	/**
	 * Get the position at the end of the game, variants will be ignored!
	 * @return the position
	 */
	public Position getPosition(){
		//Create empty default position
		Position result = new Position();
		PartyNode node = root;
		
		//Add all moves
		while(node.hasPV()){
			Move move = node.getPV().getMove();
			node = node.getPV();
			result = move.applyMoveTo(result); 
		}
		
		return result;
	}
	
	/**
	 * Get a position in the game, with endPly being the last ply before the position that
	 * should be created.
	 * @param endPly the last ply
	 * @return the position
	 */
	public Position getPosition(int endPly){
//		Create empty default position
		Position result = new Position();
		PartyNode node = root;
		
		int i = -1;
		//Add all moves
		while(node.hasPV() && (++i < endPly)){
			Move move = node.getPV().getMove();
			node = node.getPV();
			result = move.applyMoveTo(result);
		}
		
		return result;
	}
	
	//TODO Create getPosition(int, int) with variations

	public String getHTMLNotation(){
		String notation = MessageFormat.format(Resources.getMessages().getString("notationScheme"), whitePlayer, blackPlayer, date, site, tournament, ECO, startingComment);
		Position p = new Position();
		
		boolean white = true, cancelled = false, hilight = false;
		
		PartyNode node = root;
		if(root.hasPV())
			node = root.getPV();
		else return notation;
		int i = -1;
		
		while(true){
			++i;
	
			if(i % 2 == 0) white = true;
			else white = false;
			Move move = node.getMove();
			
			if(white){
				notation += "<tr><td>"+((i/2)+1)+"</td><td>";
				
				
				notation += move.getNotation(p);
				p = move.applyMoveTo(p);
				if(node.getComment() != null){
					notation += "</tr><tr><td colspan='3'>"
					+node.getComment()+"</td></tr><tr>";
					cancelled = true;
				}
			}
			else{
				if(cancelled) notation += "<tr><td>"+((i/2)+1)+"</td><td>...</td>";
				else notation += "<td bgcolor='#e3e3e3'>";
				
				notation += move.getNotation(p);
				p = move.applyMoveTo(p);
				if(node.getComment() != null){
					notation += "</tr><tr><td colspan='3'>"
					+node.getComment()+"</td></tr><tr>";
				}
				cancelled = false;
			}
			
//			notation += moves.get(i).getMove().getNotation() + " ";
//			if(!moves.get(i).getCommentAfter().trim().equals(""))
//				notation += "</tr><tr><td colspan = '2'>"+moves.get(i).getCommentAfter()+"</td></tr><tr>";
			
			if(i % 2 != 0)
				notation += "</td></tr>";
			
			white = !white;
			if(node.hasPV())
				node = node.getPV();
			else
				break;
		}
		notation += "</table></html>";
		
		
		return notation;
	}
	
	/**
	 * Adds a comment after the last move
	 * @param comment the comment to add
	 */
	public void addComment(String comment){
		lastNode.setComment(comment);
	}
	
	/**
	 * Parse a PGN notation
	 * @param pgnnotation
	 */
	public void parsePGN(String pgnnotation){
		//Append missing last character (deleted in BaseView)
		if(pgnnotation.endsWith("1-"))
			pgnnotation += "0";
		else if(pgnnotation.endsWith("0-"))
			pgnnotation += "1";
		else if(pgnnotation.endsWith("1/2-1/"))
			pgnnotation += "2";
		
		
		//Clear this party
		root = new PartyNode();
		lastNode = root;
		
		whitePlayer = "";
		blackPlayer = "";
		result = "";
		date = "";
		site = "";
		tournament = "";
		round = "";
		
		//Parse meta information
		for(String data: pgnnotation.split("\n+")){
			if(data.matches("\\s*\\[White \".+\"\\s*\\]")){
				setWhitePlayer(data.split("\"")[1]);
			}
			else if(data.matches("\\s*\\[Black \".+\"\\s*\\]")){
				setBlackPlayer(data.split("\"")[1]);
			}
			else if(data.matches("\\s*\\[Result \".+\"\\s*\\]")){
				setResult(data.split("\"")[1]);
			}
			else if(data.matches("\\s*\\[Site \".+\"\\s*\\]")){
				setSite(data.split("\"")[1]);
			}
			else if(data.matches("\\s*\\[Date \".+\"\\s*\\]")){
				setDate(data.split("\"")[1]);
			}
			else if(data.matches("\\s*\\[Event \".+\"\\s*\\]")){
				setTournament(data.split("\"")[1]);
			}
			else if(data.matches("\\s*\\[ECO \".+\"\\s*\\]")){
				setECO(data.split("\"")[1]);
			}
			else if(data.matches("\\s*\\[.*\\]\\s*")){
				additionalMetaData.add(data);
			}
		}
		
		//Find the break between meta data and the party, this consist of one or
		//more free lines!
		String[] parts = pgnnotation.split("\n\n+");
		if(parts.length != 2){
//			System.out.println("Party#parsePGN(String):No valid PGN party data found!");
			try {
				File file = File.createTempFile("pgnerr", ".log");
				ListFile.addToFile(file.getAbsolutePath(), pgnnotation);
//				System.out.println("Party#parsePGN(String):Invalid PGN dumped to "+file.getAbsolutePath());
			} catch (IOException e) {
//				System.out.println("Party#parsePGN(String):Invalid PGN dump failed!");
//				System.out.println("Party#parsePGN(String):Caused by "+e.getStackTrace());
			}
			return;
		}
		
//		System.out.println("Party#parsePGN(String):Starting to parse notation");
		
		String move = "";
        String comment = "";
        int variantLevel = 0;
		boolean isComment = false;
		
		for(int i = 0; i < parts[1].length(); ++i){
			char working = parts[1].charAt(i);
			
//			System.out.println("Move parser: Variant level is "+variantLevel+", comment: "+(isComment ? "yes": "no")
//					+", char is "+working);
			if(Character.isWhitespace(working) && !move.equals("")){
				addMove(new Move(move, getPosition()));
				move = "";
			}
			else if(working == '{'){
				isComment = true;
				if(!move.matches("\\s*"))
					addMove(new Move(move, getPosition()));
				move = "";
			}
			else if(working == '}'){
				isComment = false;
				if(!comment.matches("\\s+")){
					addComment(comment);
					comment = "";
				}
			}
			else if(working == '('){
				++variantLevel;
				if(!move.matches("\\s*"))
					addMove(new Move(move, getPosition()));
			}
			else if(working == ')'){
				--variantLevel;
				move = "";
			}
			else{
				if(variantLevel == 0 && !isComment){
					move += working;
				}
				else if(isComment && variantLevel == 0){
					comment += working;
				}
			}
		}
	}
	
	/**
	 * This will generate a PGN from this party
	 * @return a PGN string that can be saved into a PGN file
	 */
	public String generatePGN(){
		StringBuffer result = new StringBuffer();
		
		//Generate meta data header
		if(!tournament.equals(""))
			result.append( "[Event \""+tournament+"\"]\n");
		if(!site.equals(""))
			result.append("[Site \""+site+"\"]\n");
		if(!date.equals(""))
			result.append("[Date \""+date+"\"]\n");
		if(!round.equals(""))
			result.append("[Round \""+round+"\"]\n");
		result.append("[White \""+whitePlayer+"\"]\n");
		result.append("[Black \""+blackPlayer+"\"]\n");
		if(!this.result.equals(""))
			result.append("[Result \""+this.result+"\"]\n");
		if(!ECO.equals(""))
			result.append("[ECO \""+ECO+"\"]\n");
		for(String data: additionalMetaData)
			result.append(data+"\n");
		
		result.append("\n");
		//Create party body
		int counter = 0;
		Position position = new Position();
		
		
		
		
//		for(MoveData moveData: moves){
//			if(!moveData.getCommentBefore().matches("\\s+"))
//				result.append(" { "+moveData.getCommentBefore()+" } ");
//			
//			String move = moveData.getMove().getNotation(position);
//			result.append(move);
//			moveData.getMove().applyMoveTo(position);
//			
//			++counter;
//		}
		
		
		
		return result.toString();
	}
	
	public String getBlackPlayer() {
		return blackPlayer;
	}

	public String getWhitePlayer() {
		return whitePlayer;
	}

	public void setBlackPlayer(String blackPlayer) {
		this.blackPlayer = blackPlayer;
	}

	public void setWhitePlayer(String whitePlayer) {
		this.whitePlayer = whitePlayer;
	}

	/**
	 * Gets the last move of the PV
	 * @return
	 */
	public Move getLastMove() {
		if(lastNode != root)
			return lastNode.getMove();
		return Move.NO_NEW_MOVE;
	}
	
	private boolean engineNeedsToResync = false;
	public boolean needResync(){
		if(engineNeedsToResync){
			engineNeedsToResync = false;
			return true;
		}
		return false;
	}

	public String getResult() {
		return result;
	}

	public void setResult(String result) {
		this.result = result;
	}

	public String getDate() {
		return date;
	}

	public String getSite() {
		return site;
	}

	public String getTournament() {
		return tournament;
	}

	public void setDate(String date) {
		this.date = date;
	}

	public void setSite(String site) {
		this.site = site;
	}

	public void setTournament(String tournament) {
		this.tournament = tournament;
	}

	public String getRound() {
		return round;
	}

	public void setRound(String round) {
		this.round = round;
	}

	public String getECO() {
		return ECO;
	}

	public void setECO(String eco) {
		ECO = eco;
	}

	public int getPlyCount() {
		return plyCount;
	}

}
