/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.chess.party;

import klaus.chess.Move;

import java.util.Vector;

/**
 * Represents a node in a Party tree
 * @author Stephan Hillebrand
 * @since 2009-04-04 alpha3
 */
public class PartyNode {
	
	public static int LAST_ID = 0;
	
	/**
	 * A unique, auto-generated ID
	 */
	private int id = ++PartyNode.LAST_ID;
	
	private Move lastMove = Move.NO_NEW_MOVE;
	/**
	 * A vector containing all possible next moves ( that is, all variants including the pv)
	 */
	private Vector<PartyNode> nextMoves = new Vector<PartyNode>();
	
	/**
	 * The PV's (=principal variation's) index in nextMoves
	 */
	private int pv = -1;
	
	/**
	 * A comment
	 */
	private String comment = null;
	
	/**
	 * Create a new PartyNode for a new game (= root node)
	 *
	 */
	public PartyNode(){}
	
	/**
	 * Create a new Party node within a game (= normal node)
	 * @param lastMove the move transforming the parent node's position to this node's position
	 */
	public PartyNode(Move lastMove){
		this.lastMove = lastMove;
	}
	
	/**
	 * Add a new variant
	 * @param next the node after the next variant
	 */
	public void addVariant(PartyNode next){
		nextMoves.add(next);
	}
	
	/**
	 * Adds a new variant and makes it PV
	 * @param next
	 */
	public void addPV(PartyNode next){
		for(int i = 0; i < nextMoves.size(); ++i){
			//Check if the new PV is already contained as a normal variant
			if(nextMoves.get(i).equals(next)){
				pv = i;
				return;
			}
		}
		nextMoves.add(next);
		pv = nextMoves.size() - 1;
	}
	
	public boolean hasPV(){
		return pv >= 0;
	}
	
	public PartyNode getPV(){
		return nextMoves.get(pv);
	}
	
	public Move getMove(){
		return lastMove;
	}

	public int getId() {
		return id;
	}

	/**
	 * Get the comment after this move. By default, null means no comment at all is displayed, while ""
	 * or any blank string will mean an empty line might be displayed depending on how notation is handled.
	 * @return the comment string if any or null if no comment is given
	 */
	public String getComment() {
		return comment;
	}

	/**
	 * Set the comment after the move
	 * @param comment the comment if any or null in order to erase an existing comment.
	 */
	public void setComment(String comment) {
		this.comment = comment;
	}
}
