/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.components;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Toolkit;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.util.Vector;

import javax.swing.JPanel;

import klaus.Settings;
import klaus.backend.ChessBackend;
import klaus.backend.MoveListener;
import klaus.backend.ResultListener;
import klaus.backend.ScrollingBackend;
import klaus.chess.Move;
import klaus.chess.Position;
import klaus.chess.Result;
import klaus.chess.party.Party;
import klaus.dialogs.PromotionDialog;

/**
 * This component draws a chess board (what a surprise!).
 * 
 * @author Stephan Hillebrand
 * @since 2009-03-11
 */
public class ChessBoard extends JPanel {

	/**
	 * The backend used by this board
	 */
	private ChessBackend backend = null;

	/**
	 * Do not draw the piece at this X position (drag'n'drop)
	 */
	private int dndXPOS = -1;

	/**
	 * Do not draw the piece at this Y position (drag'n'drop)
	 */
	private int dndYPOS = -1;

	/**
	 * Position inside the piece (X) (drag'n'drop)
	 */
	private int dndXOffset = 0;

	/**
	 * Position inside the piece (Y) (drag'n'drop)
	 */
	private int dndYOffset = 0;

	private int size = 0;

	private int xposition = 0;

	private int yposition = 0;

	private double fieldsize = 0.0;
	
	private Font coordsFont = null;
	
	private Result result = null;
	
	/**
	 * Is the board turned? False will mean the white pieces are at the bottom of the board,
	 * true means the opposite. So this should be true whenever the player has the black pieces.
	 */
	private boolean turned = false;

	public ChessBoard(ChessBackend chessBackend) {

		this.backend = chessBackend;

		backend.addMoveListener(new MoveListener() {
			public void moveDone(Move move) {
                if(move == Move.NO_NEW_MOVE){
                    System.out.println("MoveListener#moveDone(Move) in ChessBoard#ChessBoard(ChessBackend): Caught" +
                            " Move.NO_NEW_MOVE");
                    return;
                }
				System.out.println("ChessBoard: MoveListener fired with move "+move.toXEngString());
                backend.addMove(move);
				//backend.getParty().addMove(move);
				repaint();
				//backend.setPosition(backend.getParty().getPosition());

			}
		});

		setPreferredSize(new Dimension(600, 700));
		setMinimumSize(new Dimension(200, 200));

		addMouseListener(new MouseAdapter() {
			@Override
			public void mouseReleased(MouseEvent e) {

				// Get mouse position
				int mouseX = e.getX();
				int mouseY = e.getY();

				// get field
				int fieldX = getSquareX(mouseX);

				int fieldY = getSquareY(mouseY);
				//Check if drag'n'dropp is enabled
				if(dndXPOS < 0 || dndYPOS < 0) return;
				if (fieldX >= 0 && fieldY >= 0 && fieldX < 8 && fieldY < 8
				// Check if anything moved
						&& (fieldX != dndXPOS || fieldY != dndYPOS)) {

					// Move done! Handle it.
					Move move = new Move(dndXPOS, dndYPOS, fieldX, fieldY);
					// System.out.println(move.getNotation());
					if ((backend.getParty().getPosition().getPieceAt(dndXPOS, dndYPOS) == 'P' && fieldY == 0)
							|| (backend.getParty().getPosition()
									.getPieceAt(dndXPOS, dndYPOS) == 'p')
							&& fieldY == 7) {

						move.setPromotion(PromotionDialog
								.showPromotionDialog(backend.getParty().getPosition()
										.isWhitesTurn()));

					}

					if (backend.addMove(move)) {
						repaint();

						fireListeners();
						// backend.setPosition(party.getPosition());
					} else {
						Toolkit.getDefaultToolkit().beep();
					}
				}

				dndXPOS = -1;
				dndYPOS = -1;

				repaint();
			}

			@Override
			public void mousePressed(MouseEvent e) {
				// Get mouse position
				int mouseX = e.getX();
				int mouseY = e.getY();

				// get field
				int fieldX = getSquareX(mouseX);

				int fieldY = getSquareY(mouseY);
				System.out.println("Coords dragged: "+fieldX+";"+fieldY);
				if (fieldX >= 0 && fieldY >= 0 && fieldX < 8 && fieldY < 8) {
					// Enable dnd
					dndXPOS = fieldX;
					dndYPOS = fieldY;

					// calculate dnd coords
					dndXOffset = mouseX - getXOnScreen(dndXPOS);
					dndYOffset = mouseY - getYOnScreen(dndYPOS);
//					dndX = getSquareX(dndXPOS);
//					dndY = getSquareY(dndYPOS);

				}

				repaint();
			}

			
		});
		
		addMouseWheelListener(new MouseWheelListener(){
			public void mouseWheelMoved(MouseWheelEvent e) {

				
				ScrollingBackend scroll;
				if(backend instanceof ScrollingBackend) scroll = (ScrollingBackend) backend;
				else return;
					
				if(e.getWheelRotation() < 0){

					for(int i = 0; i > e.getWheelRotation(); --i){
						scroll.goToPrevious();
					}
				}
				else if(e.getWheelRotation() > 0){
					for(int i = 0; i < e.getWheelRotation(); ++i){
						scroll.goToNext();
					}
				}
			}

		});


		addMouseMotionListener(new MouseMotionListener() {

			public void mouseDragged(@SuppressWarnings("unused")
			MouseEvent arg0) {
				repaint();
			}

			public void mouseMoved(@SuppressWarnings("unused")
			MouseEvent arg0) {
				repaint();
			}
		});

		addComponentListener(new ComponentListener() {

			public void componentHidden(@SuppressWarnings("unused")
			ComponentEvent arg0) {
				calculateSize();
			}

			public void componentMoved(@SuppressWarnings("unused")
			ComponentEvent arg0) {
				calculateSize();
			}

			public void componentResized(@SuppressWarnings("unused")
			ComponentEvent arg0) {
				calculateSize();
			}

			public void componentShown(@SuppressWarnings("unused")
			ComponentEvent arg0) {
				calculateSize();
			}

		});
	
		backend.addResultListener(new ResultListener(){
			public void result(Result res) {
                System.out.println("ChessBoard#ChessBoard(): ResultListener fired!");
				result = res;
				repaint();
			}
		});
	}
	
	/**
	 * Convert the mouse coordinates to the coords of a square (X coord)
	 */
	private int getSquareX(int mouseX){
		if(!turned)
			return (int) ((mouseX - (xposition + fieldsize)) / fieldsize);
		return (int) (8 - ((mouseX - fieldsize - xposition) / fieldsize));
	}
	/**
	 * Convert the mouse coordinates to the coords of a square (Y coord)
	 */
	private int getSquareY(int mouseY){
		if(!turned)
			return (int) ((mouseY - (yposition + fieldsize)) / fieldsize);
		return (int) (8 - ((mouseY - fieldsize - yposition) / fieldsize));
	}
	
	private int getXOnScreen(int squareX){
		if(!turned)
			return (int)(xposition + ((squareX+1)*fieldsize));
		else
			return (int)(xposition + ((8-squareX)*fieldsize));
	}
	
	private int getYOnScreen(int squareY){
		if(!turned)
			return (int)(yposition + ((squareY+1)*fieldsize));
		else
			return (int)(yposition + ((8-squareY)*fieldsize));
	}

	private void calculateSize() {
		// calculate optimal board size
		int xmaxsize = getWidth();
		int ymaxsize = getHeight();
		size = 0;
		xposition = 0;
		yposition = 0;

		if (ymaxsize >= xmaxsize) {
			size = xmaxsize;
			yposition = (ymaxsize - size) / 2;
		} else {

			size = ymaxsize;
			xposition = (xmaxsize - size) / 2;
		}
		fieldsize = size / 10.0;
		
		//determine maximum font size of coordinates
		int fontSize = 60;
		coordsFont = new Font(Font.SANS_SERIF, Font.BOLD, fontSize);
		while (true){
			/*
			 * coordinate font size should be 0.8 times (80%) the border's height
			 */
			if(getGraphics().getFontMetrics(coordsFont).getHeight() <= fieldsize * 0.8 || fieldsize < 0){
				System.out.println("Found correct font size: "+fontSize);
				return;
			}
			--fontSize;
			coordsFont = new Font(Font.SANS_SERIF, Font.PLAIN, fontSize);
		}
	}

	/**
	 * Paint the board
	 */
	@Override
	public void paintComponent(Graphics g) {
		
		if(g instanceof Graphics2D){
//			((Graphics2D) g).setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
		}
		
		String layout = Settings.getSettings().getLayoutName();
		// get the position using the party data
		Position position = backend.getPosition();
		// Draw background!
		BufferedImage bgimage = ImageCache.getImage(layout, "desk");
		for(int y = 0; y <= getHeight() / bgimage.getHeight(); y++){
			for(int x = 0; x <= getWidth() / bgimage.getWidth(); x++){
				g.drawImage(bgimage, x*bgimage.getWidth(), y*bgimage.getHeight(),null);
			}
		}
		
		// calculate optimal board size
		int xmaxsize = getWidth();
		int ymaxsize = getHeight();
		int size = 0;
		int xposition = 0;
		int yposition = 0;
		
		if(ymaxsize >= xmaxsize){
			size = xmaxsize;
			yposition = (ymaxsize - size) / 2;			
		}else{
			
			size = ymaxsize;
			xposition = (xmaxsize - size) / 2;
		}
		
		double fieldsize = size / 10.0;
		// Draw the board
		g.drawImage(ImageCache.getImage(layout, "board"), xposition, yposition, size, size, null);
		
		//Check if font size has already been calculated and do so if necessary
		if(coordsFont == null)
			calculateSize();
		
		//Print board coordinates
		int coordsLowerYPosition = (int)(yposition + (9.9*fieldsize));
		int coordsUpperYPosition = (int)(yposition + 0.9*fieldsize);
		g.setFont(coordsFont);
		String[] lineCoords = null;
		String[] rowCoords = null;
		if(!turned){
			lineCoords = new String[]{"A", "B", "C", "D", "E", "F", "G", "H"};
			rowCoords = new String[]{"8", "7", "6", "5", "4", "3", "2", "1"};
		}
		else{
			lineCoords = new String[]{"H", "G", "F", "E", "D", "C", "B", "A"};
			rowCoords = new String[]{"1", "2", "3", "4", "5", "6", "7", "8"};
		}
		
		for(int i = 0; i < 8; i++){
			int xpos = (int) (xposition + (i+1)*fieldsize + 
					(0.5*fieldsize - 0.5*g.getFontMetrics(coordsFont)
							.getStringBounds(lineCoords[i], g).getWidth()));
			
			g.drawString(lineCoords[i], xpos, coordsUpperYPosition);
			g.drawString(lineCoords[i], xpos, coordsLowerYPosition);
			
			int rowOffset = (int) g.getFontMetrics().getStringBounds(rowCoords[i], g).getWidth() / 2;
			//int height = (int) g.getFontMetrics().getHeight();
			
			g.drawString(rowCoords[i], (int)(xposition + 0.5*fieldsize -0.5*rowOffset), 
					(int)(yposition + (i+1.9)*fieldsize));
			g.drawString(rowCoords[i], (int)(xposition + 9.5*fieldsize -0.5*rowOffset), 
					(int)(yposition + (i+1.9)*fieldsize));
		}
		
		
		for(int line = 0; line < 8; line++){
			for(int row = 0; row < 8; row++){
				
				// Ignore dnd piece
				if(row == dndYPOS && line == dndXPOS)
					continue;
				if(position.getPieceAt(line, row) == ' ') continue;
				
				int fieldx, fieldy;
				if(!turned){
					fieldx = (int)(xposition + ((line+1)*fieldsize));
					fieldy = (int)(yposition + ((row+1)*fieldsize));
				}else{
					fieldx = (int)(xposition + ((8-line)*fieldsize));
					fieldy = (int)(yposition + ((8-row)*fieldsize));
				}
				String iname = pieceToFileName(position.getPieceAt(line, row));

				g.drawImage(ImageCache.getImage(layout, iname), fieldx, fieldy, (int)fieldsize, (int)fieldsize, null);
			}
		}
		
		// check if dnd active
		if(dndXPOS >= 0 && dndXPOS < 8 && dndYPOS >= 0 && dndYPOS < 8){
			// draw dnd piece
			int fieldx = (int)(getMousePosition().getX() - dndXOffset);
			int fieldy = (int)(getMousePosition().getY() - dndYOffset);
			
			String fname = pieceToFileName(position.getPieceAt(dndXPOS, dndYPOS));
			
			g.drawImage(ImageCache.getImage(layout, fname), fieldx, fieldy, (int)fieldsize, (int)fieldsize, null);
		}else{
			
			// Highlight last move
			Move m = getBackend().getParty().getLastMove();
			if(m != null){
				g.setColor(Color.GREEN);
				for(int i = 0; i < 3; i++){
					if(!turned)
						g.drawRect(
								(int)(xposition + ((m.getToX()+1)*fieldsize) +i),
								(int)(yposition + ((m.getToY()+1)*fieldsize) +i), 
								(int)(fieldsize-2*i),
								(int)(fieldsize-2*i));
					else
						g.drawRect(
								(int)(xposition + ((8-m.getToX())*fieldsize) +i),
								(int)(yposition + ((8-m.getToY())*fieldsize) +i), 
								(int)(fieldsize-2*i),
								(int)(fieldsize-2*i));
				}
			}
		}
		
		if(result == Result.whiteCheckmated || result == Result.blackCheckmated){

			g.drawImage(ImageCache.getImage("../../share/klaus/graphics/gray.png"), 0, 0, getWidth(), getHeight(), null);

			//draw checkmate image
			g.drawImage(ImageCache.getImage("../../share/klaus/graphics/playing/checkmate.png"), xposition, yposition,
					size, size, null);
		}
		else if(result == Result.blackTimeout || result == Result.whiteTimeout){
			g.drawImage(ImageCache.getImage("../../share/klaus/graphics/gray.png"), 0, 0, getWidth(), getHeight(), null);
			g.drawImage(ImageCache.getImage("../../share/klaus/graphics/playing/timeout.png"), xposition, yposition,
					size, size, null);
		}
	}

	private String pieceToFileName(char piece) {
		String iname = "";
		switch (piece) {
		case 'p':
			iname = "black_pawn";
			break;
		case 'P':
			iname = "white_pawn";
			break;
		case 'n':
			iname = "black_knight";
			break;
		case 'N':
			iname = "white_knight";
			break;
		case 'b':
			iname = "black_bishop";
			break;
		case 'B':
			iname = "white_bishop";
			break;
		case 'r':
			iname = "black_rook";
			break;
		case 'R':
			iname = "white_rook";
			break;
		case 'q':
			iname = "black_queen";
			break;
		case 'Q':
			iname = "white_queen";
			break;
		case 'k':
			iname = "black_king";
			break;
		case 'K':
			iname = "white_king";
			break;
		}
		return iname;
	}

	public ChessBackend getBackend() {
		return backend;
	}

	public Party getParty() {
		return backend.getParty();
	}

	public void setBackend(ChessBackend backend) {
		this.backend = backend;
	}

	/*public void setParty(Party party) {
		backend.setPosition(party.getPosition());
	}*/

	private Vector<ChessBoardListener> listeners = new Vector<ChessBoardListener>();

	public void addChessBoardListener(ChessBoardListener listener) {
		listeners.add(listener);
	}

	public void removeChessBoardListener(ChessBoardListener listener) {
		listeners.remove(listener);
	}

	public void fireListeners() {
		for (ChessBoardListener listener : listeners) {
			listener.gameChanged(backend.getParty());
		}
	}
	
	public boolean isTurned(){
		return turned;
	}
	
	public void setTurned(boolean turned){
		this.turned = turned;
		repaint();
	}
}
