/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.engine;

import klaus.Settings;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;


/**
 * Saves the data for a chess engine
 * @author Stephan Hillebrand
 * @since 2009-03-15 alpha2
 */
public class EngineData {
	
	/**
	 * Shows that no engine was loaded.
	 */
	public static final EngineData NO_ENGINE = new EngineData(0,"none","none");
	
	
	/**
	 * The engine's command line, that is the path to an exe file (with settings?) on Windows
	 * or to a shwll script/binary on unixoid systems.
	 * Windows engines can be run on Linux etc. using "wine C:\\Program\ Files\\Your\\Engines\\name\\engine.exe"
	 * or something of that sort if wine is installed (Wine is available for MacOS X, Solaris/OpenSolaris,
	 * Linux and Net/FreeBSD, but *BSD does not support Java afaik).
	 */
	private String executable = null;
	
	private String name = null;
	
	/**
	 * The engine's type / protocol
	 */
	private int engineType = 0;
	
	/**
	 * UCI Engine (not supported up to now)
	 */
	public static final int ENGINE_PROTOCOL_UCI = 1;
	
	/**
	 * WinBoard / XBoard engine. Winboard and XBoard are quite similar with the difference
	 * that a WinBoard engine is a Windows executable while a XBoard engine is a Linux/Solaris/... binary
	 * file. This is handled by Java so no differences have to be paid attention to.
	 */
	public static final int ENGINE_PROTOCOL_WINBOARD_XBOARD = 2;
	
	/**
	 * A hash map containing engine settings. As settings depend on engine in UCI mode,
	 * this has to be done in a hash map.
	 */
	private HashMap<String, String> properties = new HashMap<String, String>();
	
	/**
	 * Create the data for a new engine.
	 * @param type the engine's type/protocol (Winboard/Xboard/UCI???)
	 * @param cmdLine the engine's command line (e.g. /usr/bin/gnuchessx)
	 */
	public EngineData(int type, String cmdLine, String name){
		engineType = type;
		executable = cmdLine;
		this.name = name;
	}

	/**
	 * Create an engine backend from this engine data
	 * @return the engine backend
     * @throws IOException if the engine was not started due to an IO Exception
	 */
	public Engine createEngine() throws IOException{
		if(engineType == ENGINE_PROTOCOL_WINBOARD_XBOARD){
			//Create a WinBoard/XBoadr engine
			WinBoardXBoardEngine engine = null;
			//try {
				engine = new WinBoardXBoardEngine(executable);
			//} catch (IOException e) {
			//	System.err.println("Failed to create engine: ");
			//	e.printStackTrace();
			//}
			return engine;
		}
		else if(engineType == ENGINE_PROTOCOL_UCI){
            UCIEngine engine = null;
            engine = new UCIEngine(executable);
            return engine;
        }
        return null;
		
	}
	
	/**
	 * Create a description of the engine, e. g. for the engine configuration view
	 * @return a short description of the engine
	 */
	public String getDescription(){
		String res =  "<html><table><tr><td rowspan=\"3\">"
		
		+(amIDefault() ? "<img src=\""+new File("graphics/default.png").getAbsolutePath()+"\"></td>"
				:"<img src=\""+new File("graphics/default-no.png").getAbsolutePath()+"\"></td>")
		
		+"<td><b>"+name + " ("+ (engineType == ENGINE_PROTOCOL_UCI ? "UCI": "WinBoard/XBoard")+ 
		")</b></td></tr><tr><td>Executable: "+executable+"</td></table></html>";
		return res;
	}

	@Override
	public String toString(){
		return getDescription();
	}
	
	public int getEngineType() {
		return engineType;
	}

	public String getExecutable() {
		return executable;
	}

	public String getName() {
		return name;
	}

	public void setEngineType(int engineType) {
		this.engineType = engineType;
	}

	public void setExecutable(String executable) {
		this.executable = executable;
	}

	public void setName(String name) {
		this.name = name;
	}
	
	/**
	 * Checks if this is the default engine
	 * @return true if this is the default engine, otherwise false
	 */
	public boolean amIDefault(){
		if(Settings.getSettings().getDefaultEngine().equals(getName()))
			return true;
		return false;
	}
	
	/**
	 * Edit an engine setting
	 * @param key the settings's key
	 * @param value the setting's value
	 */
	public void setSetting(String key, String value){
		properties.put(key, value);
	}
	
	public String getSetting(String key){
		return properties.get(key);
	}
	
	public HashMap<String, String> getSettings(){
		return properties;
	}
	
}
