/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.engine;

import klaus.backend.MoveListener;
import klaus.chess.Move;
import klaus.chess.Position;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.Queue;
import java.util.Vector;

/**
 *
 * This class is used for binding an UCI engine (e. g. Toga, Fruit) to Klaus.
 */
public class UCIEngine extends Engine implements Runnable{

    private Process connection;
    private BufferedReader input;
    private PrintWriter output;
    private Thread engineThread;

    /**
     * Flag for shutting down the engine
     */
    private boolean running = true;

    /**
     * Check if anything can be sent to the engine
     */
    private boolean blocked = true;

    private int wtime = 0, btime = 0, wtimeUsed = 0, bTimeUsed = 0, increment = 0, movesToGo = 0;

    private Vector<String> messageStack = new Vector<String>();

    //Position position = new Position();

    /**
     * <p>Create and start a new UCI engine. This method is almost the same as in WinBoardXBoardEngine
     * since actually both are started the same way.</p>
     *
     * <p>Still, there are some differences: Unlike XBoard engines, UCI engines require some
     * initialization stuff, this is done in the run() method.</p>
     * @param commandLine the command line (consisting of an executable's path and (optionally) arguments
     * passed to the engine). Note: If there are spaces or some special characters in the path, "" will be
     * required.
     * @throws IOException if the engine could not be started correctly for some reason (file not found,
     * crash during startup)
     */
    public UCIEngine (String commandLine) throws IOException{
        System.out.println("Engine:                 Trying to establish a pipe to \""+commandLine+"\"...");

		String[] info = commandLine.split("\"\\s+", 2);
		if(info.length == 1)
			info = new String[]{info[0], ""};
		else if(info.length == 0) return;
		System.out.println("Engine:                Command is "+info[0]+", args: "+info[1]);
		connection = new ProcessBuilder(info[0], info[1]).start();
		System.out.println("Engine:                 O. K., opening streams...");
		input = new BufferedReader(new InputStreamReader(connection.getInputStream()));
		//Flush after each output line
		output = new PrintWriter(connection.getOutputStream(), true);

		engineThread = new Thread(this);
		engineThread.start();

        Thread clockThread = new Thread(){
            @Override
            public void run() {
                while(true){
                    updateTime();
                    yield();
                }
            }
        };
        clockThread.start();
    }


    public void addMove(Move m){
        //TODO enter pondering here!
        super.addMove(m);
        //Update time so that the switching is handled correctly
        updateTime();
    }

    @Override
    public void startCalculation() {
        //Start calculation
        sendMessage("position fen "+party.getPosition().toFEN(), false);

        int time = 0;

        sendMessage("go wtime "+(wtime/1000)+" btime "+(btime/1000), false);
    }

    @Override
    public void cancel() {
        forceMove();
    }

    @Override
    public void unload() {
        //Stop engine thread
        running = false;

        //The engine thread will do anything else needed, like terminating the engine process.
    }

    @Override
    public void forceMove() {
        sendMessage("stop", true);
    }

    @Override
    public synchronized void setTimeModel(int timeInMinutes, int moves, int increment) {
        wtime = (timeInMinutes*60000) - wtimeUsed;
        btime = (timeInMinutes*60000) - bTimeUsed;
        this.increment = increment;
        movesToGo = moves - party.getPosition().getNoOfMove();
        lastUpdateTime = System.currentTimeMillis();

        //TODO Need increment and periods here
    }

    /**
     * System time at last update
     */
    private long lastUpdateTime = System.currentTimeMillis();
    /**
     * Updates clocks, needs to be called exactly once a second.
     *
     */
    private synchronized void updateTime(){

        long difference = System.currentTimeMillis() - lastUpdateTime;
        lastUpdateTime = System.currentTimeMillis();

        if(party.getPosition().isWhitesTurn()){
            wtime -= difference;
            wtimeUsed = (int) (wtimeUsed + difference);
        }
        else{
            btime -= difference;
            bTimeUsed += difference;
        }
    }

    /**
     * Send a string to the engine.
     * @param message the message, command, whatsoever
     * @param force force sending the message if blocked is true? This should only be set to true by run()!
     */
    public synchronized void sendMessage(final String message, boolean force){
        System.out.println("UCIEngine sending message: "+message);
        if((!blocked) ||force){
            //send output directly
            output.println(message);
            output.flush();
        }
        else{
            messageStack.add(message);
        }
    }




    private synchronized void lockEngine(){
        blocked = true;
    }

    private synchronized void unlockEngine(){
        //Handle stack
        while(messageStack.size() != 0){
            sendMessage(messageStack.get(0), true);
            messageStack.remove(0);
        }
        //unblock
        blocked = true;
    }

    /**
     * Handle an engine command
     * @param message a string sent by the engine
     */
    private void handle(String message){
        message = message.trim();
        String lowerCaseMessage = message.toLowerCase();

        if(lowerCaseMessage.startsWith("bestmove")){

            //The engine stopped thinking and gave a move to the interface. This move is handled here.

            String[] data = message.split(" +");

            Move m = new Move(data[1], party.getPosition());
            addMove(m);
            fireMoveListeners(m);
            //TODO Save pondering result!
        }

        if(lowerCaseMessage.equals("readyok")){
            //Engine is ready for playing, unlock transmission
            blocked = false;
        }
    }

    public void run() {
        //Set up the engine using the UCI command
        sendMessage("uci", true);
        sendMessage("isready", true);
        
        while(running){
            try{
				String text = input.readLine();
				if(text != null&&!text.equals(""))
					handle(text);

				Thread.sleep(50);
			}
            catch(InterruptedException ie){}
			catch(IOException ioe){
				System.err.println("Engine:                IOException in Engine handling: ");
				ioe.printStackTrace();

				System.out.println("Engine closing...");
				connection = null;
				engineThread = null;
                running = false;
			}


            //This will cause the thread to wait until the engine terminates.
          //  System.out.println("Engine stopped with exit code "+connection.exitValue());
          //  System.out.println("Engine thread halted.");

        }
        System.out.println("Engine stopping, please wait.");

        sendMessage("quit", true);
    }
}
