/*
 * Klaus chess program
 * Copyright (C) 2009 Stephan Hillebrand <stefh123@yahoo.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package klaus.engine;

import klaus.chess.Move;
import klaus.chess.Position;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;

/**
 * Connects to a WinBoard/XBoard engine.
 * @author Stephan Hillebrand
 * @since 2009-03-15 alpha2
 */
public class WinBoardXBoardEngine extends Engine implements Runnable {
	private Process connection = null;
	private BufferedReader input = null;
	private PrintWriter output = null;
	private Thread engineThread = null;
	private boolean cancel = false;
	private boolean running = true;
	private boolean inForceMode = false;
	
	private static final char[] lines = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'};
	private static final char[] rows = {'8', '7', '6', '5', '4', '3', '2', '1'};
	
	
	public WinBoardXBoardEngine(String commandLine) throws IOException{
		System.out.println("Engine:                 Trying to establish a pipe to \""+commandLine+"\"...");
		
		String[] info = commandLine.split("\"\\s+", 2);
		if(info.length == 1)
			info = new String[]{info[0], ""};
		else if(info.length == 0) return;
		System.out.println("Engine:                Command is "+info[0]+", args: "+info[1]);
		connection = new ProcessBuilder(info[0], info[1]).start();
		System.out.println("Engine:                 O. K., opening streams...");
		input = new BufferedReader(new InputStreamReader(connection.getInputStream()));
		//Flush after each output line
		output = new PrintWriter(connection.getOutputStream(), true);
		
		engineThread = new Thread(this);
		engineThread.start();
	}
	
		
	private synchronized void handle(String message){
		System.out.println("Engine: Receive "+message);
		String orig = message + "";
		message = message.toLowerCase().trim();
		if(message.startsWith("my move is: ")||message.startsWith("move ")){
            //Get the last part of the engine message, this is the move.
            String[] cmd = message.trim().split(" ");
			String move = cmd[cmd.length-1];
			System.out.println("Engine:               Adding move "+move);
			if(!cancel){
				Move m = new Move(move, party.getPosition());
				fireMoveListeners(m);
				clearDebuggingListeners();
				party.addMove(m);
			}
		}
		if(message.matches("\\d+.*\\d+\\s+\\d+\\s+\\d+\\s+.+")){
			//This seems to be a debug information!
			String[] parts = orig.trim().split("\\s+", 5);
			int ply = Integer.parseInt(parts[0].replaceAll("\\D+", ""));
			String score = parts[1];
			int time = Integer.parseInt(parts[2]);
			int nodes = Integer.parseInt(parts[3]);
			String pv = parts[4];
			
			DebuggingInformation info = new DebuggingInformation(ply, score, time, nodes, pv);
			fireDebuggingListeners(info);
		}
	}
	
	public void run(){
		//Setup XBoard 2
		sendmsg("xboard");
		sendmsg("protover 2");
		//Init new game
		sendmsg("new");
		//Set random on (will make games more interesting in GNUChess)
		sendmsg("random");
		sendmsg("post");
		while(running){
			try{
				String text = input.readLine();
				if(text != null&&!text.equals(""))
					handle(text);
				
				Thread.sleep(50);
			}catch(InterruptedException ie){
				
			}catch(IOException ioe){
				System.err.println("Engine:                IOException in Engine handling: ");
				ioe.printStackTrace();
				
				System.out.println("Engine closing...");
				connection = null;
				engineThread = null;
			}
		}
	}
	
	private void resyncPosition(){
		Position position = party.getPosition();
		sendmsg("new");
		setForceMode(true);
		
		if(!position.isWhitesTurn()) sendmsg("a2a3");
		sendmsg("edit");
		sendmsg("#");
		
		boolean color = true;
		for(int x = 0; x < 8; x++){
			for(int y = 0; y < 8; y++){
				if(position.getPieceAt(x, y) == ' ') continue;
				
				if(Character.isUpperCase(position.getPieceAt(x, y)) != color)
					sendmsg("c");
					
				sendmsg(Character.toUpperCase(position.getPieceAt(x, y))+""+lines[x]+""+rows[y]);
				
			}
		}
		sendmsg(".");
		
	}
	
	private void setForceMode(boolean mode){
		if(!mode){
			inForceMode = true;
			sendmsg("go");
		}
		else{
			inForceMode = false;
			sendmsg("force");
		}
	}
	
	private synchronized void sendmsg(String message){
		System.out.println("Engine: Send           "+message);
		output.println(message);
	}

	@Override
	public void startCalculation() {
		cancel = false;
		
		if(party.needResync()){
			resyncPosition();
			sendmsg("go");
		}
		
		System.out.println("Engine:               starting to think");
		sendmsg(party.getLastMove().toXEngString());
		if(inForceMode){
			setForceMode(false);
		}
	}
	

	@Override
	public void cancel() {
		cancel = true;
	}

	@Override
	public void unload() {
		System.out.println("Engine:                Unloading");
		//Tell the engine to quit
		sendmsg("quit");
		
		//Shut down communication thread
		running = false;
	}


	@Override
	public void forceMove() {
		sendmsg("?");
	}

    @Override
    public void setTimeModel(int timeInMinutes, int movesToNextPeriod, int increment) {
        System.out.println("Time model set, timeInMinutes is "+timeInMinutes);
        String toEngine = "level "+movesToNextPeriod+" "+timeInMinutes+" "+ increment;
        sendmsg(toEngine);
    }
}
