package klaus.views;

import klaus.Resources;
import klaus.Settings;
import klaus.backend.MoveListener;
import klaus.backend.OfflinePlayingBackend;
import klaus.backend.ResultListener;
import klaus.chess.Move;
import klaus.chess.Result;
import klaus.chess.party.Party;
import klaus.components.*;
import klaus.components.error.ErrorDialog;
import klaus.dialogs.JokeCategory;
import klaus.dialogs.Jokes;
import klaus.dialogs.NewGameModel;
import klaus.engine.Engine;
import klaus.engine.EngineData;
import myutils.components.StatusBar;

import javax.swing.*;
import java.awt.*;
import java.io.IOException;
import java.text.MessageFormat;import java.util.ResourceBundle;


/**
 * This is a view for playing.
 * @author Stephan Hillebrand
 *
 */
public class PlayingView extends View implements ChessView {
	
	/**
	 * A panel for showing the game's notation
	 */
	private NotationViewer notation = null;
	
	private String whiteEngineName = "";
	private String blackEngineName = "";
    private Engine whiteEngine = null, blackEngine = null;
	
	private PlayingToolbar pt;
	private EngineDebuggingPanel debug;
	private OfflinePlayingBackend backend = null;
	private ChessClock chessClock = null;
	
	private StatusBar status = new StatusBar(new String[]{
			"                                                      "
	});
	
	/**
	 * The chess board
	 */
	private ChessBoard board = null;
	private ResourceBundle resourceBundle = Resources.getMessages();
	public PlayingView() {
		//Prevent the frame size from changing when pack() is called
		setPreferredSize(getSize());
		
		//A border layout is needed.
		setLayout(new BorderLayout());
		
		//Create the tool bar
		pt = new PlayingToolbar(this);
		add(pt, BorderLayout.NORTH);
		
		//First of all, a new splitPane for right-to-left splitting
		//is created. The right half will take the notation while the right
		//pane is splitted again.
		final JSplitPane rightleft = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
		
		//create the Notation Viewer and add it
		notation = new NotationViewer();
		/*
		 * Set Notation pane to empty
		 */
		notation.refresh(new Party());
//		rightleft.setRightComponent(notation);
		
		//then another split is done. The bottom area is used for an engine information frame,
		//the top area is the main board area.
		final JSplitPane clockNotation = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		
		//Create the backend
		backend = new OfflinePlayingBackend();
		//Initialize the engine(s)
		
		//create and add the board
		board = new ChessBoard(backend);
		rightleft.setLeftComponent(board);
		
		//create the clock
		chessClock = new ChessClock(backend);
		clockNotation.setTopComponent(chessClock);
		
		JSplitPane notationToEngine = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		notationToEngine.setTopComponent(notation);
		
		
		//create the engine info area
		debug = new EngineDebuggingPanel();
		notationToEngine.setBottomComponent(debug);
		clockNotation.setBottomComponent(notationToEngine);
		
		
		
		//Add topbottom to rightleft
		rightleft.setRightComponent(clockNotation);
		
		rightleft.setDividerLocation(720);
		
		rightleft.setContinuousLayout(true);
		clockNotation.setContinuousLayout(true);
		
		add(rightleft, BorderLayout.CENTER);
		
		add(status, BorderLayout.SOUTH);
		

	
		board.addChessBoardListener(new ChessBoardListener(){

			public void gameChanged(Party p) {
				//Handle a game change -> refresh notation!
				System.out.println("Refreshing notation viewer");
				notation.refresh(p);
			}
		});
		
		//refresh notation whenever the engine moves
		backend.addMoveListener(new MoveListener() {
			public void moveDone(@SuppressWarnings("unused")
			Move move) {
				notation.refresh(board.getParty());
			}
		});
		
		backend.addResultListener(new ResultListener() {
			public void result(Result result) {
				System.out.println("PlayingView:           ResultListener received evt: "+result);
				if((result == Result.blackCheckmated || result == Result.blackTimeout)
					||
				(result == Result.whiteCheckmated || result == Result.whiteTimeout)){
					System.out.println("PlayingView:           ResultListener - Game ended!");
					
					//The engine lost!
					if((backend.getWhite() == null )==(backend.getBlack() == null)){
						//Both players or no player at all is an engine!
						//Give a neutral comment...
						
						String won = resourceBundle.getString("white.wins");
						if(result == Result.whiteCheckmated || result == Result.whiteTimeout)
							won = resourceBundle.getString("black.wins");
						else if(result == Result.stalemate) won = resourceBundle.getString("draw.stalemate");
						else if(result == Result.repetition) won = resourceBundle.getString("draw.repetition");
						else if(result == Result.fiftyMoves) won = resourceBundle.getString("draw.fifty.moves");
						
						status.setText(0, won);
						
					}else{
						if((getResult(result) == 1 && backend.getWhite() == null)
						||
						(getResult(result) == -1 && backend.getBlack() == null)){
							EventQueue.invokeLater(new Runnable(){
							public void run(){
//							//The human player won
//							JOptionPane.showMessageDialog(null, Jokes.getJokes().getRandomJoke(JokeCategory.computerLost),
//									"Sie haben gewonnen!", 
//									JOptionPane.INFORMATION_MESSAGE);
								
								status.setText(0, Jokes.getJokes().getRandomJoke(JokeCategory.computerLost));
								}});
						}else if(getResult(result) == 0){
							EventQueue.invokeLater(new Runnable(){
								public void run(){
//							JOptionPane.showMessageDialog(null, Jokes.getJokes().getRandomJoke(JokeCategory.drawn),
//									"Remis", 
//									JOptionPane.INFORMATION_MESSAGE);
									status.setText(0, Jokes.getJokes().getRandomJoke(JokeCategory.drawn));
								}});
						}
						else{
							EventQueue.invokeLater(new Runnable(){
								public void run(){
//							JOptionPane.showMessageDialog(null, Jokes.getJokes().getRandomJoke(JokeCategory.computerWon),
//									"Sie haben verloren!", 
//									JOptionPane.INFORMATION_MESSAGE);
									status.setText(0, Jokes.getJokes().getRandomJoke(JokeCategory.computerWon));
								}});
						}
					}
					
				}
			}
		});
	}
	
	
	public ChessBoard getChessBoard(){
		return board;
	}
	
	/**
	 * get the party's result as an integer 
	 * @param result the result
	 * @return -1 if black won, 0 for draw, 1 for white
	 */
	private int getResult(Result result){
		if(result == Result.blackCheckmated || result == Result.blackTimeout) return 1;
		if(result == Result.whiteCheckmated || result == Result.whiteTimeout) return -1;
		return 0;
	}
	
	public Party getParty(){
		return board.getParty();
	}
	
	@Override
	public String getDescription(){
		return MessageFormat.format(resourceBundle.getString("0.1.local.game"), getParty().getWhitePlayer(), getParty().getBlackPlayer());
	}

	public String getBlackEngineName() {
		return blackEngineName;
	}

	public String getWhiteEngineName() {
		return whiteEngineName;
	}

	public void setBlackEngineName(String blackEngineName) {
		this.blackEngineName = blackEngineName;
		
		if(blackEngineName != ""){
            blackEngine = getEngineByName(blackEngineName);
			backend.setBlackEngine(blackEngine);
			debug.addEngine(backend.getBlack());
		}else
			backend.setBlackEngine(null);
	}

	public void setWhiteEngineName(String whiteEngineName) {
		this.whiteEngineName = whiteEngineName;
		
		if(whiteEngineName != ""){
            whiteEngine = getEngineByName(whiteEngineName);
			backend.setWhiteEngine(whiteEngine);
			debug.addEngine(backend.getWhite());
		}else
			backend.setWhiteEngine(null);
	}

	private Engine getEngineByName(String engineName){
		for(EngineData engine: Settings.getSettings().getEngines()){
			if(engine.getName().equals(engineName))
                try{
                    return engine.createEngine();
                } catch(IOException ioe){
                    ErrorDialog.showErrorDialog(ioe, false);
                }

		}
		System.err.println("Engine not found: "+engineName+" (in OflinePlayingView#getEngineByName(String)");
		return null;
	}

	public OfflinePlayingBackend getBackend() {
		return backend;
	}
	
	@Override
	public boolean closeView() {
		if(backend.isGameRunning()){
			int select = JOptionPane.showConfirmDialog(
					this,
                    resourceBundle.getString("party.still.running"),
                    resourceBundle.getString("really.close"),
					JOptionPane.YES_NO_OPTION,
					JOptionPane.WARNING_MESSAGE
			);
			if(select == JOptionPane.NO_OPTION)
				return false;
		}
		chessClock.stop();
        //Stop the engines (if any)
        if(blackEngine != null) blackEngine.unload();
        if(whiteEngine != null) whiteEngine.unload();
		return super.closeView();
	}
	/**
	 * set the game model for clock setup
	 * @param model
	 */
	public void setGameModel(NewGameModel model) {
		if(board.getBackend() instanceof OfflinePlayingBackend)
			((OfflinePlayingBackend) board.getBackend()).setGameModel(model);
	}

	
	
}
